import unittest
from hashlib import md5

import requests

from tira.tira_redirects import mirror_url, redirects


def md5_of_first_kilobyte_of_http_resource(url):
    if not url.startswith("https://files.webis.de") and not url.startswith("https://zenodo.org/records"):
        raise ValueError(f"URL {url} is not from webis.de respectively zenodo")
    return md5(requests.get(url, headers={"Range": "bytes=0-1024"}).content).hexdigest()


class TestRedirects(unittest.TestCase):
    def test_redirect_ir_lab_leipzig_wise23_train_inputs(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/training-20231104-inputs.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-/training-20231104-training.zip"

        expected_md5_first_kilobyte = "6e2e445a72e282c25c545154b427d3fd"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_train_truths(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/training-20231104-truth.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-truth/training-20231104-training.zip"
        expected_md5_first_kilobyte = "6d1ef9e09c40e9b0664366a2a003fa64"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_padua_sose23_train_inputs(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/training-20231104-inputs.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-/longeval-tiny-train-20240315-training.zip"

        expected_md5_first_kilobyte = "6e2e445a72e282c25c545154b427d3fd"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_padua_sose23_train_truths(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/training-20231104-truth.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-truth/longeval-tiny-train-20240315-training.zip"
        expected_md5_first_kilobyte = "6d1ef9e09c40e9b0664366a2a003fa64"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_jena_inputs(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/jena-topics-small-20240119-inputs.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-/jena-topics-small-20240119-training.zip"

        expected_md5_first_kilobyte = "b013dcb175648f0190fb5b5ab9f8cc5b"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_jena_truths(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/jena-topics-small-20240119-truth.zip?download=1"
        dataset_url = "https://www.tira.io/data-download/training/input-truth/jena-topics-small-20240119-training.zip"
        expected_md5_first_kilobyte = "aa25f6931bed856fdb29d0374ff107e3"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_validation_inputs(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/validation-20231104-inputs.zip?download=1"
        expected_md5_first_kilobyte = "1ad750ebde6828971ec63cda1d415be6"
        dataset_url = "https://www.tira.io/data-download/training/input-/validation-20231104-training.zip"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_leipzig_inputs(self):
        expected_redirect = (
            "https://zenodo.org/records/10628882/files/leipzig-topics-small-20240119-inputs.zip?download=1"
        )
        dataset_url = "https://www.tira.io/data-download/training/input-/leipzig-topics-small-20240119-training.zip"

        expected_md5_first_kilobyte = "09a02fdfb03c9797842ff7adc453f850"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_leipzig_truths(self):
        expected_redirect = (
            "https://zenodo.org/records/10628882/files/leipzig-topics-small-20240119-truth.zip?download=1"
        )
        dataset_url = (
            "https://www.tira.io/data-download/training/input-truth/leipzig-topics-small-20240119-training.zip"
        )
        expected_md5_first_kilobyte = "7a60ab76ed8c3423bd407932a1e7e910"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_wise23_validation_truths(self):
        expected_redirect = "https://zenodo.org/records/10628882/files/validation-20231104-truth.zip?download=1"
        expected_md5_first_kilobyte = "92ba3a87c1d4cceebccbefd1aa5f944e"
        dataset_url = "https://www.tira.io/data-download/training/input-truth/validation-20231104-training.zip"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        self.assertEqual(expected_redirect, actual_redirect)
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_sose23_inputs(self):
        expected_redirect = (
            "https://zenodo.org/records/10628640/files/iranthology-20230618-training-inputs.zip?download=1"
        )
        expected_md5_first_kilobyte = "f0727e93741e50b32270883e821075d0"
        dataset_url = "https://www.tira.io/data-download/training/input-/iranthology-20230618-training.zip"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        self.assertEqual(expected_redirect, actual_redirect)
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)

    def test_redirect_ir_lab_leipzig_sose23_truths(self):
        expected_redirect = (
            "https://zenodo.org/records/10628640/files/iranthology-20230618-training-truths.zip?download=1"
        )
        expected_md5_first_kilobyte = "7dd9d2cb48c3bc8515fd0d5c973c5fd8"
        dataset_url = "https://www.tira.io/data-download/training/input-truth/iranthology-20230618-training.zip"

        actual_redirect = redirects(url=dataset_url)["urls"][0]
        actual_mirror = mirror_url(actual_redirect)
        actual_md5_first_kilobyte = md5_of_first_kilobyte_of_http_resource(actual_redirect)

        self.assertEqual(expected_redirect, actual_redirect)
        self.assertEqual(expected_redirect, actual_mirror)
        self.assertEqual(expected_md5_first_kilobyte, actual_md5_first_kilobyte)



import unittest

from tira.check_format import check_format

from . import EMPTY_OUTPUT, IR_QUERY_OUTPUT, VALID_RUN_OUTPUT


class TestCheckFormatForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        with self.assertRaises(Exception):
            check_format(EMPTY_OUTPUT, "does-not-exist")

    def test_invalid_validator_on_query_output(self):
        with self.assertRaises(Exception):
            check_format(IR_QUERY_OUTPUT, "does-not-exist")

    def test_invalid_validator_on_valid_run_output_output(self):
        with self.assertRaises(Exception):
            check_format(VALID_RUN_OUTPUT, "does-not-exist")

    def test_multiple_invalid_validators_on_empty_output(self):
        with self.assertRaises(Exception):
            check_format(EMPTY_OUTPUT, ["d1", "d2"])

    def test_multiple_invalid_validators_on_query_output(self):
        with self.assertRaises(Exception):
            check_format(IR_QUERY_OUTPUT, ["d1", "d2"])

    def test_multiple_invalid_validators_on_valid_run_output_output(self):
        with self.assertRaises(Exception):
            check_format(VALID_RUN_OUTPUT, ["d1", "d2"])



from pathlib import Path

from tira.check_format import FormatMsgType

_ERROR = FormatMsgType.ERROR
_OK = FormatMsgType.OK
_WARN = FormatMsgType.WARN

RESOURCES = Path(__file__).parent.parent / "resources"
VALID_RUN_OUTPUT = RESOURCES / "ranking-outputs"
RUN_OUTPUT_WITH_TOO_FEW_QUERIES = RESOURCES / "ranking-output-invalid-too-few-queries"
RUN_OUTPUT_WITH_DUPLICATE_DOCUMENTS = RESOURCES / "ranking-output-invalid-duplicate-documents"
RUN_OUTPUT_WITH_TOO_FEW_COLUMNS = RESOURCES / "ranking-output-invalid-too-few-columns"
EMPTY_OUTPUT = RESOURCES / "input-run-01" / "1"
IR_QUERY_OUTPUT = RESOURCES / "query-processing-outputs" / "query-segmentation"



import json
import unittest

import pandas as pd
from approvaltests import verify_as_json

from tira.local_client import Client


def retrieval_submission(queries):
    queries = pd.DataFrame([{"qid": str(i)} for i in queries])
    tira = Client("tests/resources/")
    retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

    ret = retriever(queries)
    return [{"qid": i["qid"], "docno": i["docno"], "score": i["score"]} for _, i in ret.iterrows()]


class TestPtFromRetrieverSubmissionLoaderTest(unittest.TestCase):
    def test_for_multiple_queries(self):
        expected = [
            {"qid": "1", "docno": "doc-1", "score": 10},
            {"qid": "1", "docno": "doc-2", "score": 9},
            {"qid": "3", "docno": "doc-3", "score": 1},
        ]

        actual = retrieval_submission([1, 3])

        self.assertEqual(len(actual), 3)
        self.assertEqual(actual[0], expected[0])
        self.assertEqual(actual[1], expected[1])
        self.assertEqual(actual[2], expected[2])

    def test_cascading_with_retriever_is_idempotent_01(self):
        queries = pd.DataFrame([{"qid": str(i)} for i in [1]])
        tira = Client("tests/resources/")
        retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

        ret = (retriever >> retriever).search(queries)
        ret = [json.loads((i.to_json())) for _, i in ret.iterrows()]
        verify_as_json(ret)

    def test_cascading_with_retriever_is_idempotent_02(self):
        queries = pd.DataFrame([{"qid": str(i)} for i in [1]])
        tira = Client("tests/resources/")
        retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

        ret = (retriever >> retriever >> retriever >> retriever).search(queries)
        ret = [json.loads((i.to_json())) for _, i in ret.iterrows()]
        verify_as_json(ret)

    def test_cascading_with_retriever_is_idempotent_03(self):
        queries = pd.DataFrame([{"qid": str(i)} for i in [1]])
        tira = Client("tests/resources/")
        retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

        ret = (
            (retriever >> retriever >> retriever >> retriever) >> (retriever >> retriever >> retriever >> retriever)
        ).search(queries)
        ret = [json.loads((i.to_json())) for _, i in ret.iterrows()]
        verify_as_json(ret)

    def test_feature_union_from_retriever_01(self):
        queries = pd.DataFrame([{"qid": str(i)} for i in [1]])
        tira = Client("tests/resources/")
        retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

        ret = (retriever >> (retriever**retriever)).search(queries)
        ret = [json.loads((i.to_json())) for _, i in ret.iterrows()]
        verify_as_json(ret)

    def test_feature_union_from_retriever_02(self):
        queries = pd.DataFrame([{"qid": str(i)} for i in [1]])
        tira = Client("tests/resources/")
        retriever = tira.pt.from_retriever_submission("ir-benchmarks/tira-ir-starters/retriever", dataset="d1")

        ret = (retriever >> (retriever**retriever**retriever**retriever)).search(queries)
        ret = [json.loads((i.to_json())) for _, i in ret.iterrows()]
        verify_as_json(ret)

    def test_for_single_query_no_1(self):
        expected = [
            {"qid": "1", "docno": "doc-1", "score": 10},
            {"qid": "1", "docno": "doc-2", "score": 9},
        ]

        actual = retrieval_submission([1])

        self.assertEqual(len(actual), 2)
        self.assertEqual(actual[0], expected[0])
        self.assertEqual(actual[1], expected[1])

    def test_for_single_query_no_3(self):
        expected = [
            {"qid": "3", "docno": "doc-3", "score": 1},
        ]

        actual = retrieval_submission([3])

        self.assertEqual(len(actual), 1)
        self.assertEqual(actual[0], expected[0])

    def test_retrieval_submission_from_rest_api(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.from_submission(
            "ir-benchmarks/tira-ir-starter/BM25 Re-Rank (tira-ir-starter-pyterrier)",
            pt.get_dataset("irds:disks45/nocr/trec-robust-2004"),
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "306"}]))), 1000)
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[0].to_dict()["docno"], "LA021790-0114")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[0].to_dict()["qid"], "306")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[999].to_dict()["docno"], "FBIS4-47956")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[999].to_dict()["qid"], "306")

    def test_retrieval_submission_from_rest_api_different_id(self):
        import pyterrier as pt

        from tira.rest_api_client import Client
        from tira.third_party_integrations import ensure_pyterrier_is_loaded

        ensure_pyterrier_is_loaded()
        tira = Client()

        q = tira.pt.from_submission(
            "ir-benchmarks/tira-ir-starter/BM25 Re-Rank (tira-ir-starter-pyterrier)",
            pt.get_dataset("irds:ir-benchmarks/disks45-nocr-trec-robust-2004-20230209-training"),
        )
        self.assertEqual(len(q(pd.DataFrame([{"qid": "306"}]))), 1000)
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[0].to_dict()["docno"], "LA021790-0114")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[0].to_dict()["qid"], "306")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[999].to_dict()["docno"], "FBIS4-47956")
        self.assertEqual(q(pd.DataFrame([{"qid": "306"}])).iloc[999].to_dict()["qid"], "306")



#!/usr/bin/env python3
import argparse
import os
import subprocess
import sys


def parse_args():
    parser = argparse.ArgumentParser(prog="tira-run-notebook", description="")

    parser.add_argument(
        "--input",
        type=str,
        help="The directory that contains the input data (within tira-run this should be '$inputDataset').",
        required=True,
    )
    parser.add_argument("--notebook", type=str, help="The notebook to execute.", required=True)
    parser.add_argument(
        "--output",
        type=str,
        help="The directory for storing output data (within tira-run this should be '$outputDir').",
        required=True,
    )

    return parser.parse_args()


def main():
    args = parse_args()
    os.environ["TIRA_INPUT_DIRECTORY"] = args.input
    os.environ["TIRA_OUTPUT_DIRECTORY"] = args.output

    command = f"runnb --allow-not-trusted {args.notebook}"
    try:
        return_code = subprocess.check_call(command, shell=True, stdout=sys.stdout, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        return_code = 2
    sys.exit(return_code)



from abc import ABC
from typing import TYPE_CHECKING, Union, overload

if TYPE_CHECKING:
    import io
    from typing import Any, Dict, Optional

    import pandas as pd

# .. todo:: this file needs further documentation


class TiraClient(ABC):
    def all_datasets(self) -> "pd.DataFrame":
        pass

    def all_softwares(self) -> "pd.DataFrame":
        pass

    def print_overview_of_all_software(self) -> None:
        pass

    def all_evaluators(self) -> "pd.DataFrame":
        pass

    def all_evaluated_appraoches(self) -> "pd.DataFrame":
        pass

    def docker_software() -> "Any":
        # .. todo:: typehint
        pass

    def run_was_already_executed_on_dataset(self, approach, dataset) -> bool:
        # .. todo:: typehint
        pass

    def get_run_output(self, approach, dataset, allow_without_evaluation=False) -> str:
        # .. todo:: typehint
        pass

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None) -> "Optional[Dict[str, str]]":
        # .. todo:: typehint
        pass

    def docker_registry(self):
        return "registry.webis.de"

    def modify_task(self, task_id: str, to_rename: "Dict[str, Any]"):
        # .. todo:: typehint
        pass

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "tuple[pd.DataFrame, str]":
        # .. todo:: typehint
        ...

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "pd.DataFrame":
        # .. todo:: typehint
        ...

    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "Union[pd.DataFrame, tuple[pd.DataFrame, str]]":
        # .. todo:: typehint
        pass

    def create_new_upload(self, task_id: str, vm_id: str) -> "Optional[str]":
        """
        Creates a new upload and returns the newly created id. Returns None on failure.
        """
        pass

    def submit_run(self, task_id: str, vm_id: str, dataset_id: str, upload_id: str, run: "io.IOBase") -> bool:
        """
        Submits the runfile located at `run` for the given task and vm for the given upload id. Returns true on success,
        false  otherwise.
        """
        pass

    def __extract_dataset_identifier(self, dataset: any):
        """Extract the dataset identifier from a passed object.

        Args:
            dataset (any): Some representation of dataset.

        Returns:
            Optional[dict]: The dataset identifier if available.
        """
        if hasattr(dataset, "irds_ref"):
            return self.__extract_dataset_identifier(dataset.irds_ref())
        if hasattr(dataset, "dataset_id"):
            return dataset.dataset_id()
        return dataset

    def __matching_dataset(self, datasets, dataset_identifier) -> "Optional[dict]":
        """Find the dataset identified by the passed dataset_identifier in all passed datasets.

        Args:
            datasets (List[dict]): TIRA representations of datasets.
            dataset_identifier (_type_): identifier of the dataset to be found.

        Returns:
            Optional[dict]: The TIRA representation of the dataset if found in the datasets.
        """
        datasets = [i for i in datasets if "id" in i and len(i["id"]) > 2]

        for dataset in datasets:
            if dataset_identifier and dataset["id"] == dataset_identifier:
                return dataset

        if len(dataset_identifier.split("/")) == 2:
            task_identifier, dataset_in_task = dataset_identifier.split("/")

            for dataset in datasets:
                if "default_task" not in dataset or not dataset["default_task"]:
                    continue

                if not task_identifier or not dataset_in_task:
                    continue

                if task_identifier == dataset["default_task"] and dataset_in_task == dataset["id"]:
                    return dataset

        for dataset in datasets:
            if "ir_datasets_id" not in dataset or not dataset["ir_datasets_id"] or len(dataset["ir_datasets_id"]) < 3:
                continue

            if dataset_identifier and dataset_identifier == dataset["ir_datasets_id"]:
                return dataset



import os
import unittest
from pathlib import Path

from tira.profiling_integration import ProfilingIntegration
from tira.rest_api_client import Client


def mocked_tira_client(run_dir):
    class MockedTiraClient:
        def get_run_output(self, approach, dataset, allow_without_evaluation=False):
            return f"{Path(os.path.abspath(__file__)).parent}/{run_dir}"

    return MockedTiraClient()


class ParsingOfProfilingTest(unittest.TestCase):
    def test_error_message_is_thrown_if_no_tira_client_is_available(self):
        tira_client = None
        profiling_integration = ProfilingIntegration(tira_client)
        with self.assertRaises(Exception) as context:
            profiling_integration.from_submission("run_id", "dataset")
        print(context.exception)
        self.assertTrue("No profiling data available for approach" in str(context.exception))

    def test_error_message_is_thrown_if_profile_data_is_available(self):
        tira_client = mocked_tira_client(".")
        profiling_integration = ProfilingIntegration(tira_client)
        with self.assertRaises(Exception) as context:
            profiling_integration.from_submission("run_id", "dataset")
        print(context.exception)
        self.assertTrue("No profiling data available for run" in str(context.exception))

    def test_profiling_info_available_as_dict(self):
        tira_client = mocked_tira_client("/resources/profiling_logs/a")
        profiling_integration = ProfilingIntegration(tira_client)
        expected = {"timestamp": 0.0, "key": "ps_cpu", "value": 0.2}
        actual = profiling_integration.from_submission("run_id", "dataset")

        self.assertEqual(expected, actual[0])

    def test_profiling_info_has_elpsed_time_as_last_slot(self):
        tira_client = mocked_tira_client("/resources/profiling_logs/b")
        profiling_integration = ProfilingIntegration(tira_client)
        expected = {"timestamp": 2.0, "key": "elapsed_time", "value": 2}
        actual = profiling_integration.from_submission("run_id", "dataset")

        self.assertEqual(expected, actual[-1])

    def test_profiling_info_available_as_pd(self):
        tira_client = mocked_tira_client("/resources/profiling_logs/c")
        profiling_integration = ProfilingIntegration(tira_client)
        expected = {"timestamp": 0.0, "key": "ps_cpu", "value": 0.2}
        actual = profiling_integration.from_submission("run_id", "dataset", return_pd=True)

        self.assertEqual(expected, actual.iloc[0].to_dict())

    def test_profiling_info_has_elpsed_time_as_last_slot_from_zip(self):
        tira_client = mocked_tira_client("/resources/profiling_logs2/d")
        profiling_integration = ProfilingIntegration(tira_client)
        expected = {"timestamp": 64.0, "key": "elapsed_time", "value": 64}
        actual = profiling_integration.from_submission("run_id", "dataset")

        self.assertEqual(expected, actual[-1])

    def test_prod_system(self):
        expected = {"timestamp": 38.0, "key": "elapsed_time", "value": 38}
        actual = Client().profiling.from_submission(
            "reneuir-2024/tinyfsu/tiny-fsu-bert", "dl-top-1000-docs-20240701-training"
        )

        self.assertEqual(expected, actual[-1])



#!/usr/bin/env python
import argparse
import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

from tira.io_utils import huggingface_model_mounts
from tira.local_client import Client
from tira.local_execution_integration import LocalExecutionIntegration
from tira.rest_api_client import Client as RestClient
from tira.third_party_integrations import extract_previous_stages_from_docker_image

if TYPE_CHECKING:
    from .tira_client import TiraClient


def setup_upload_run_command(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("runfile", type=argparse.FileType("rb"), help="Path to the runfile to be uploaded")
    parser.add_argument("--dataset-id", required=True, type=str)
    parser.add_argument("--task-id", required=False, default=os.environ.get("TIRA_TASK_ID"), type=str)
    parser.add_argument("--vm-id", required=False, default=os.environ.get("TIRA_VM_ID"), type=str)
    parser.add_argument("--upload-group", required=True, type=str)

    # Register the upload_run_command method as the "main"-method
    parser.set_defaults(executable=upload_run_command)


def upload_run_command(args: argparse.Namespace) -> int:
    # TODO: allow the user to override some settings (e.g., API key and URL) via arguments
    tira = RestClient()
    if args.task_id is None:
        raise argparse.ArgumentTypeError("Please populate --task-id or set the environment variable TIRA_TASK_ID")
    if args.vm_id is None:
        raise argparse.ArgumentTypeError("Please populate --vm-id or set the environment variable TIRA_VM_ID")
    upload_id = tira.create_upload_group(args.task_id, args.vm_id, args.upload_group)
    success = tira.upload_run(
        task_id=args.task_id, vm_id=args.vm_id, dataset_id=args.dataset_id, upload_id=upload_id, filestream=args.runfile
    )
    return 0 if success else 1


def parse_args():
    parser = argparse.ArgumentParser(prog="tira-run")
    parser.add_argument("--input-directory", required=False, default=str(os.path.abspath(".")))
    parser.add_argument("--input-dataset", required=False, default=None)
    parser.add_argument("--input-run", required=False, default=None)
    parser.add_argument("--input-run-directory", required=False, default=None)
    # Not required if the subcommand "tira-run submit" is used.
    # FIXME: the arguments in this group should probably be subcommands
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("--image")
    group.add_argument("--approach")
    group.add_argument("--export-dataset", required=False, default=None, type=str)
    group.add_argument("--export-approach-output", nargs="*", required=False, default=None, type=str)
    group.add_argument("--export-submission-from-jupyter-notebook", required=False, default=None, type=str)
    group.add_argument("--export-submission-environment", nargs="*", required=False, default=None, type=str)
    parser.add_argument(
        "--command",
        required=False,
        help=(
            "Command to execute in the container. Has access to variables $inputDataset (points to the directory with"
            " the inputs), $outputDir (results should be persisted in this directory), and optionally $inputRun (points"
            " to the directory with the outputs of the previous stage)."
        ),
    )
    parser.add_argument("--verbose", required=False, default=False, action="store_true")
    parser.add_argument(
        "--gpus",
        required=False,
        default=False,
        type=str,
        help='GPU devices to add to the container ("all" or -1 to pass all GPUs, or the number of devices).',
    )
    parser.add_argument("--evaluate", required=False, default=False, type=bool)
    parser.add_argument("--evaluation-directory", required=False, default=str(os.path.abspath("tira-evaluation")))
    parser.add_argument("--dry-run", required=False, default=False, type=bool)
    parser.add_argument("--dataset-for-export-approach-output", required=False, default=False, type=str)
    parser.add_argument(
        "-v",
        action="append",
        default=None,
        help="Additional volume mounts as <host-path>:<container-path>:<read-write-mode>",
        required=False,
    )
    parser.add_argument(
        "--allow-network",
        required=False,
        default=False,
        type=bool,
        help=(
            "Is the network available during the execution? (Default = False to improve reproducibility, containers"
            " that import ir-datasets can have access to the network, they can be tested with setting allow-network to"
            " true.)"
        ),
    )
    parser.add_argument("--output-directory", required=False, default=str(os.path.abspath("tira-output")))

    parser.add_argument("--push", required=False, default="false")
    parser.add_argument(
        "--skip-local-test",
        required=False,
        default=False,
        action="store_true",
        help="Skip the execution of the local test, only upload the submission to TIRA.",
    )
    parser.add_argument(
        "--tira-docker-registry-token", required=False, default=os.environ.get("TIRA_DOCKER_REGISTRY_TOKEN")
    )
    parser.add_argument(
        "--tira-docker-registry-user", required=False, default=os.environ.get("TIRA_DOCKER_REGISTRY_USER")
    )
    parser.add_argument("--tira-client-token", required=False, default=os.environ.get("TIRA_CLIENT_TOKEN"))
    parser.add_argument("--tira-client-user", required=False, default=os.environ.get("TIRA_CLIENT_USER"))
    parser.add_argument("--tira-vm-id", required=False, default=os.environ.get("TIRA_VM_ID"))
    parser.add_argument("--tira-task-id", required=False, default=os.environ.get("TIRA_TASK_ID"))
    parser.add_argument("--tira-code-repository-id", required=False, default=os.environ.get("TIRA_CODE_REPOSITORY_ID"))
    parser.add_argument(
        "--fail-if-output-is-empty",
        required=False,
        default=False,
        action="store_true",
        help="After the execution of the software, fail if the output directory is empty.",
    )
    parser.add_argument(
        "--mount-hf-model",
        nargs="+",
        default=[],
        help=(
            "Mount models from the local huggingface hub cache (i.e., $HOME/.cache/huggingface/hub) into the container"
            " during execution. This is intended to remove redundancy so that the models must not be embedded into the"
            " Docker images."
        ),
    )

    subparsers = parser.add_subparsers()
    setup_upload_run_command(subparsers.add_parser("upload", help="For uploading runs"))

    args = parser.parse_args()
    if args.export_submission_from_jupyter_notebook:
        return args

    if args.export_submission_environment:
        return args

    if os.path.exists(args.output_directory) and len(os.listdir(args.output_directory)) > 0:
        parser.error(
            f"The output directory {os.path.abspath(args.output_directory)} is not empty. Please empty it or provide a"
            " new output directory with --ouptut-directory."
        )

    if (args.image is None) == (args.approach is None) == (args.export_dataset):
        parser.error("You have to exclusively use either --approach or --image.")
    if (args.image is None) != (args.command is None):
        args.command = LocalExecutionIntegration().extract_entrypoint(args.image)

        if args.command is not None:
            print(f'Use command from Docker image "{args.command}".')
        else:
            parser.error(
                "I could not find a command to execute, please either configure the entrypoint of the image or use"
                " --command."
            )

    args.previous_stages = [] if not args.input_run else [args.input_run]

    if args.input_run is None and args.input_run_directory is None and args.image is not None:
        args.input_run = extract_previous_stages_from_docker_image(args.image, args.command)
        args.command = LocalExecutionIntegration().make_command_absolute(args.image, args.command)
        args.previous_stages = args.input_run
        if args.input_run and len(args.input_run) == 1:
            args.input_run = args.input_run[0]

    if args.push.lower() == "true":
        print("I will check that the API key and the credentials for the image registry are valid.")
        rest_client = RestClient()
        docker_authenticated = rest_client.local_execution.docker_client_is_authenticated()

        api_key_valid = rest_client.api_key_is_valid()

        if args.tira_client_token:
            if api_key_valid:
                args.tira_client_token = rest_client.api_key
            elif "TIRA_CLIENT_TOKEN" in os.environ and os.environ.get("TIRA_CLIENT_TOKEN"):
                rest_client = RestClient(api_key=os.environ.get("TIRA_CLIENT_TOKEN"))
                api_key_valid = rest_client.api_key_is_valid()
                if api_key_valid:
                    args.tira_client_token = rest_client.api_key
                else:
                    parser.error(
                        "The option --tira-client-token (or environment variable TIRA_CLIENT_TOKEN) is required when"
                        " --push is active."
                    )
            else:
                rest_client = RestClient(api_key=args.tira_client_token)
                if not rest_client.api_key_is_valid():
                    parser.error(
                        "The option --tira-client-token (or environment variable TIRA_CLIENT_TOKEN) is required when"
                        " --push is active."
                    )

        if not args.tira_task_id and args.input_dataset:
            args.tira_task_id = args.input_dataset.split("/")[0]

        if not args.tira_task_id:
            parser.error(
                "The option --tira-task-id (or environment variable TIRA_TASK_ID) is required when --push is active."
            )

        if not args.tira_vm_id:
            tmp = rest_client.metadata_for_task(args.tira_task_id)
            if (
                tmp
                and "status" in tmp
                and 0 == tmp["status"]
                and "context" in tmp
                and "user_vms_for_task" in tmp["context"]
            ):
                if len(tmp["context"]["user_vms_for_task"]) != 1:
                    parser.error(
                        f'You have multiple vms ({tmp["context"]["user_vms_for_task"]}), use option --tira-vm-id to'
                        " specify the vm."
                    )
                else:
                    args.tira_vm_id = tmp["context"]["user_vms_for_task"][0]
            else:
                parser.error(
                    "The option --tira-vm-id (or environment variable TIRA_VM_ID) is required when --push is active."
                )

        if not args.tira_client_user:
            tmp = rest_client.metadata_for_task(args.tira_task_id)
            if tmp and "status" in tmp and 0 == tmp["status"] and "context" in tmp and "user_id" in tmp["context"]:
                args.tira_client_user = tmp["context"]["user_id"]
            else:
                parser.error(
                    "The option --tira-client-user (or environment variable TIRA_CLIENT_USER) is required when --push"
                    " is active."
                )

        if (
            not docker_authenticated
            and not args.tira_docker_registry_token
            and not rest_client.local_execution.login_docker_client(args.tira_task_id, args.tira_vm_id)
        ):
            parser.error(
                "The option --tira-docker-registry-token (or environment variable TIRA_DOCKER_REGISTRY_TOKEN) is"
                " required when --push is active."
            )

        if (
            not docker_authenticated
            and not args.tira_docker_registry_user
            and not rest_client.local_execution.login_docker_client(args.tira_task_id, args.tira_vm_id)
        ):
            parser.error(
                "The option --tira-docker-registry-user (or environment variable TIRA_DOCKER_REGISTRY_USER) is required"
                " when --push is active."
            )

        args.previous_stages = [
            rest_client.public_runs(
                args.tira_task_id, args.input_dataset.split("/")[1], i.split("/")[1], i.split("/")[2]
            )
            for i in args.previous_stages
        ]
        if "none" in args.previous_stages or None in args.previous_stages:
            parser.error(
                "One of the previous stages is not public. Please contact the organizer with a link to your code so"
                " that they can ensure it is public."
            )

    if not args.approach and not args.image:
        parser.error("Please specify what you want to run.")

    return args


def main(args=None):
    args = args if args else parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)

    client: "TiraClient" = Client()

    # Subcommands store their executable into args.executable which takes the parsed arguments and returns an integer
    if hasattr(args, "executable") and args.executable is not None:
        exit(args.executable(args))

    if args.export_submission_from_jupyter_notebook:
        ret = LocalExecutionIntegration().export_submission_from_jupyter_notebook(
            args.export_submission_from_jupyter_notebook
        )
        if not ret:
            exit(1)
            return
        print(ret)
        return

    if args.export_submission_environment:
        from tira.io_utils import all_environment_variables_for_github_action_or_fail

        for i in all_environment_variables_for_github_action_or_fail(args.export_submission_environment):
            print(i)
        return

    if args.export_dataset:
        if len(args.export_dataset.split("/")) != 2:
            print(f"Please pass arguments as: --export-dataset <task>/<tira-dataset>. Got {args.export_dataset}")
            return

        task, dataset = args.export_dataset.split("/")
        print(f'Export dataset "{dataset}" for task "{task}" to directory "{args.output_directory}".')

        tira = RestClient()
        data_dir = tira.download_dataset(task, dataset)

        if os.path.exists(args.output_directory):
            print(
                f"The directory {args.output_directory} specified by --output-directory already exists. I do not"
                " overwrite the directory, please remove it manually if you want to export the dataset."
            )
        shutil.copytree(data_dir, args.output_directory)
        return

    if args.export_approach_output:
        if not args.dataset_for_export_approach_output:
            print(
                "Please specify the dataset for which the approach outputs should be exported via"
                " --dataset-for-export-approach-output"
            )
            return

        if os.path.exists(args.output_directory):
            print(
                f"The directory {args.output_directory} specified by --output-directory already exists. I do not"
                " overwrite the directory, please remove it manually if you want to export the dataset."
            )
            return

        tira = RestClient()
        data_dirs = []

        for approach in args.export_approach_output:
            if len(approach.split("/")) != 3:
                print(
                    f"Please pass arguments as: --export-approach-output <task>/<tira-dataset>/approach. Got {approach}"
                )

        for approach in args.export_approach_output:
            data_dirs += [tira.get_run_output(approach, args.dataset_for_export_approach_output, True)]

        for num, data_dir in zip(range(len(data_dirs)), data_dirs):
            shutil.copytree(data_dir, Path(args.output_directory) / str(num + 1))

        return

    input_dir = args.input_directory
    evaluate = False
    if args.input_dataset and "none" != args.input_dataset.lower():
        if len(args.input_dataset.split("/")) != 2:
            print(f"Please pass arguments as: --input-dataset <task>/<tira-dataset>. Got {args.input_dataset}")
            return

        task, dataset = args.input_dataset.split("/")
        tira = RestClient()
        print(f"Ensure that the input dataset {dataset} is available.")
        input_dir = tira.download_dataset(task, dataset)
        print(f"Done: Dataset {dataset} is available locally.")

        if args.input_run and not isinstance(args.input_run, list) and "none" != args.input_run.lower():
            print(f"Ensure that the input run {args.input_run} is available.")
            args.input_run = tira.get_run_output(args.input_run, dataset, True)
            print("Done: input run is available locally.")
        elif args.input_run and not isinstance(args.input_run, list) and len(args.input_run) > 0:
            temp_dir = "/tmp/" + tempfile.TemporaryDirectory().name
            os.makedirs(temp_dir, exist_ok=True)
            for num, input_run in zip(range(len(args.input_run)), args.input_run):
                print(f"Ensure that the input run {input_run} is available.")
                input_run = tira.get_run_output(input_run, dataset, True)
                shutil.copytree(input_run, temp_dir + "/" + str(1 + num))
            args.input_run = temp_dir
        elif args.input_run_directory and "none" != args.input_run_directory.lower():
            args.input_run = os.path.abspath(args.input_run_directory)

        if args.evaluate:
            print(f"Ensure that the evaluation truth for dataset {dataset} is available.")
            evaluate = tira.get_configuration_of_evaluation(task, dataset)
            evaluate["truth_directory"] = tira.download_dataset(task, dataset, truth_dataset=True)
            print(f"Done: Evaluation truth for dataset {dataset} is available.")

    if args.mount_hf_model:
        hf_models = huggingface_model_mounts(args.mount_hf_model)
        hf_models = [k + ":" + v["bind"] + ":" + v["mode"] for k, v in hf_models.items()]
        print(f"The following models from huggingface are mounted: {hf_models}\n\n")
        if not args.v:
            args.v = []
        args.v += hf_models

    print(
        f"""
########################################## TIRA RUN CONFIGURATION ######################################################
# image={args.image}
# command={args.command}
########################################################################################################################
"""
    )
    gpus = 0
    if args.gpus:
        gpus = -1 if args.gpus == "all" else int(args.gpus)

    if not args.skip_local_test:
        client.local_execution.run(
            identifier=args.approach,
            image=args.image,
            command=args.command,
            input_dir=input_dir,
            output_dir=args.output_directory,
            dry_run=args.dry_run,
            allow_network=args.allow_network,
            input_run=args.input_run,
            additional_volumes=args.v,
            evaluate=evaluate,
            eval_dir=args.evaluation_directory,
            gpu_count=gpus,
        )
    else:
        print("Skip the local test execution of the software.")

    if args.fail_if_output_is_empty and (
        not os.path.exists(args.output_directory) or not os.listdir(args.output_directory)
    ):
        raise ValueError("The software produced an empty output directory, it likely failed?")

    if args.push.lower() == "true":
        print(
            f"""
######################################### Review the Outputs of your Software ##########################################
# Your software produced the following outputs in {args.output_directory}.
# Please shortly review them before we push your software.
########################################################################################################################
"""
        )
        if not os.path.exists(args.output_directory) or not os.listdir(args.output_directory):
            print("Your software did not produce any output, please review the logs above.")
        else:
            for f in os.listdir(args.output_directory):
                print(f' - {f}: {os.path.getsize(args.output_directory + "/" + f)} bytes')

        if not args.fail_if_output_is_empty:
            continue_user = input("Are the outputs correct and should I push the software to TIRA? (y/N) ").lower()

            if not continue_user or continue_user.lower() not in ["y", "yes"]:
                print("You did not specify yes, I will not push the software.")
                return

        registry_prefix = client.docker_registry() + "/code-research/tira/tira-user-" + args.tira_vm_id + "/"
        print("Push Docker image")

        image = client.local_execution.push_image(args.image, registry_prefix, args.tira_task_id, args.tira_vm_id)
        print("Upload TIRA_SOFTWARE")
        prev_stages = []
        for i in args.previous_stages:
            prev_stages += [
                str(
                    i["job_id"]["software_id"]
                    if i["job_id"]["software_id"]
                    else ("upload-" + str(i["job_id"]["upload_id"]))
                )
            ]

        tira = RestClient(api_key=args.tira_client_token, api_user_name=args.tira_client_user)
        tira.add_docker_software(
            image,
            args.command,
            args.tira_vm_id,
            args.tira_task_id,
            args.tira_code_repository_id,
            dict(os.environ),
            prev_stages,
            args.mount_hf_model,
        )



import gzip
import os
import re
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Sequence, Union


class FormatMsgType(Enum):
    OK = 0
    WARN = 1
    ERROR = 2


_fmt = FormatMsgType


class FormatBase:
    def all_lines(self, f: Path):
        try:
            f_size = f.stat().st_size
            if f_size > self.max_size():
                raise ValueError(f"File {f} is too large (size {f_size} exceeds configured {self.max_size()}).")
        except:
            raise ValueError(f"File {f} is not readable.")

        if str(f).endswith(".gz"):
            with gzip.open(f, "rt") as o:
                return o.readlines()
        else:
            with open(f, "r") as o:
                return o.readlines()

    def max_size(self):
        return 10 * 1024 * 1024


class RunFormat(FormatBase):
    """Checks if a given output is a valid run file."""

    def check_format(self, run_output: Path):
        if (run_output / "run.txt").exists() and (run_output / "run.txt.gz").exists():
            msg = f"Found multiple run.txt or run.txt.gz files: {os.listdir(run_output)} ."
            return [_fmt.ERROR, msg]
        if not (run_output / "run.txt").exists() and not (run_output / "run.txt.gz").exists():
            msg = "No file run.txt or run.txt.gz was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            return [_fmt.ERROR, msg]

        try:
            # maximum size of 10MB
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, repr(e)]

        if len(lines) < 10:
            return [_fmt.ERROR, f"The run file contains only {len(lines)} lines, this is likely an error."]

        query_to_docs = {}
        for l in lines:
            l = l.strip()
            cols = re.split("\\s+", l)
            if len(cols) != 6:
                return [
                    _fmt.ERROR,
                    f'Invalid line in the run file, expected 6 columns, but found a line "{l}" with {len(cols)} columns.',
                ]
            qid, _, docno, _, _, _ = cols
            if qid not in query_to_docs:
                query_to_docs[qid] = set()
            if docno in query_to_docs[qid]:
                return [
                    _fmt.ERROR,
                    f'The run file has duplicate documents: the document with id "{docno}" appears multiple times for query "{qid}".',
                ]
            query_to_docs[qid].add(docno)

        if len(query_to_docs.keys()) < 3:
            return [_fmt.ERROR, f"The run file has only {len(query_to_docs)} queries which is likely an error."]

        return [_fmt.OK, "The run.txt file has the correct format."]

    def all_lines(self, run_output):
        if (run_output / "run.txt").exists():
            return [i.strip() for i in super().all_lines(run_output / "run.txt")]
        if (run_output / "run.txt.gz").exists():
            return [i.strip() for i in super().all_lines(run_output / "run.txt.gz")]
        else:
            raise ValueError("Could not find a file run.txt or run.txt.gz")


class JsonlFormat(FormatBase):
    """Checks if a given output is a valid jsonl file."""

    def check_format(self, run_output: Path):
        matches = [i for i in os.listdir(run_output) if i.endswith(".jsonl")]
        if len(matches) != 1:
            msg = "No unique *.jsonl file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            return [_fmt.ERROR, msg]
        else:
            return [_fmt.OK, "The jsonl file has the correct format."]


FORMAT_TO_CHECK = {"run.txt": lambda: RunFormat(), "*.jsonl": lambda: JsonlFormat()}

SUPPORTED_FORMATS = set(sorted(list(FORMAT_TO_CHECK.keys())))


def lines_if_valid(run_output: Path, format: Union[str, Sequence[str]]):
    """Load all lines from a user file if they are valid

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
    """
    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = FORMAT_TO_CHECK[format]()
    result, msg = checker.check_format(run_output)
    if result != _fmt.OK:
        raise ValueError(msg)

    return checker.all_lines(run_output)


def check_format(run_output: Path, format: Union[str, Sequence[str]]):
    """Check if the provided run output is in the specified format. Provides debug messages intended for users.

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
    """
    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = FORMAT_TO_CHECK[format]()
    return checker.check_format(run_output)



IRDS_TO_TIREX_DATASET = {
    "clueweb09/en/trec-web-2009": "clueweb09-en-trec-web-2009-20230107-training",
    "clueweb09/en/trec-web-2010": "clueweb09-en-trec-web-2010-20230107-training",
    "clueweb09/en/trec-web-2011": "clueweb09-en-trec-web-2011-20230107-training",
    "clueweb09/en/trec-web-2012": "clueweb09-en-trec-web-2012-20230107-training",
    "clueweb12/trec-web-2013": "clueweb12-trec-web-2013-20230107-training",
    "clueweb12/trec-web-2014": "clueweb12-trec-web-2014-20230107-training",
    "disks45/nocr/trec-robust-2004": "disks45-nocr-trec-robust-2004-20230209-training",
    "antique/test": "antique-test-20230107-training",
    "argsme/2020-04-01/touche-2020-task-1": "argsme-touche-2020-task-1-20230209-training",
    "argsme/2020-04-01/touche-2021-task-1": "argsme-touche-2021-task-1-20230209-training",
    "clueweb12/touche-2020-task-2": "clueweb12-touche-2020-task-2-20230209-training",
    "clueweb12/touche-2021-task-2": "clueweb12-touche-2021-task-2-20230209-training",
    "clueweb12/b13/trec-misinfo-2019": "clueweb12-trec-misinfo-2019-20240214-training",
    "cord19/fulltext/trec-covid": "cord19-fulltext-trec-covid-20230107-training",
    "cranfield": "cranfield-20230107-training",
    "disks45/nocr/trec7": "disks45-nocr-trec7-20230209-training",
    "disks45/nocr/trec8": "disks45-nocr-trec8-20230209-training",
    "gov/trec-web-2002": "gov-trec-web-2002-20230209-training",
    "gov/trec-web-2003": "gov-trec-web-2003-20230209-training",
    "gov/trec-web-2004": "gov-trec-web-2004-20230209-training",
    "gov2/trec-tb-2004": "gov2-trec-tb-2004-20230209-training",
    "gov2/trec-tb-2005": "gov2-trec-tb-2005-20230209-training",
    "gov2/trec-tb-2006": "gov2-trec-tb-2006-20230209-training",
    "medline/2004/trec-genomics-2004": "medline-2004-trec-genomics-2004-20230107-training",
    "medline/2004/trec-genomics-2005": "medline-2004-trec-genomics-2005-20230107-training",
    "medline/2017/trec-pm-2017": "medline-2017-trec-pm-2017-20230211-training",
    "medline/2017/trec-pm-2018": "medline-2017-trec-pm-2018-20230211-training",
    "msmarco-passage/trec-dl-2019/judged": "msmarco-passage-trec-dl-2019-judged-20230107-training",
    "msmarco-passage/trec-dl-2020/judged": "msmarco-passage-trec-dl-2020-judged-20230107-training",
    "nfcorpus/test": "nfcorpus-test-20230107-training",
    "vaswani": "vaswani-20230107-training",
    "wapo/v2/trec-core-2018": "wapo-v2-trec-core-2018-20230107-training",
    "longeval/heldout": "longeval-heldout-20230513-training",
    "longeval/long-september": "longeval-long-september-20230513-training",
    "longeval/short-july": "longeval-short-july-20230513-training",
    "longeval/train": "longeval-train-20230513-training",
    "trec-tip-of-the-tongue/dev": "trec-tip-of-the-tongue-dev-20230607-training",
}

TIREX_DATASETS = IRDS_TO_TIREX_DATASET.values()



import json
import os
from glob import glob

import pandas as pd
from packaging import version

from tira.local_execution_integration import LocalExecutionIntegration
from tira.pandas_integration import PandasIntegration
from tira.pyterrier_integration import PyTerrierIntegration
from tira.rest_api_client import Client as RestClient
from tira.tira_client import TiraClient


class Client(TiraClient):
    def __init__(self, directory=".", rest_client=None):
        self.pd = PandasIntegration(self)
        self.pt = PyTerrierIntegration(self)
        self.directory = directory + "/"
        self.tira_cache_dir = os.environ.get("TIRA_CACHE_DIR", os.path.expanduser("~") + "/.tira")
        self.rest_client = rest_client if rest_client else RestClient()
        self.local_execution = LocalExecutionIntegration(self.rest_client)

    def all_datasets(self) -> pd.DataFrame:
        ret = []
        for i in glob(self.directory + "*/training-datasets/"):
            cnt = 0
            for j in glob(i + "*"):
                cnt += len(list(open(j)))

            ret += [{"dataset": i.split("/training-datasets/")[0], "records": cnt}]

        for i in set([j.split("/")[1] for j in glob(self.directory + "**/**/**/job-executed-on-*.txt")]):
            ret += [{"dataset": i}]

        return pd.DataFrame(ret).sort_values("dataset")

    def ___load_softwares(self):
        softwares = [json.loads(i) for i in open(self.directory + ".tira/submitted-software.jsonl")]

        return {i["TIRA_TASK_ID"] + "/" + i["TIRA_VM_ID"] + "/" + i["TIRA_SOFTWARE_NAME"]: i for i in softwares}

    def all_softwares(self) -> pd.DataFrame:
        ret = []
        for software_id, software_definition in self.___load_softwares().items():
            ret += [
                {
                    "approach": software_id,
                    "id": str(int(software_definition["TIRA_SOFTWARE_ID"].split("docker-software-")[1])),
                    "team": software_definition["TIRA_VM_ID"],
                    "image": software_definition["TIRA_IMAGE_TO_EXECUTE_IN_DOCKERHUB"],
                    "command": software_definition["TIRA_COMMAND_TO_EXECUTE"],
                    "ids_of_previous_stages": [str(int(i)) for i in software_definition["TIRA_IDS_OF_PREVIOUS_STAGES"]],
                }
            ]

        return pd.DataFrame(ret)

    def print_overview_of_all_software(self) -> None:
        pre_text = """# Software for [$TASK_ID](https://www.tira.io/task/$TASK_ID)

<p id="instructions">Place <code>todo-replace</code> in a directory <code>$PWD/input</code>.
Then run the software as described below to produce results in <code>$PWD/output</code>.</p>
<p>For each software three ways to run it are shown below:</p>
<ul>
<li id="description-docker"><b>Docker</b>: on the command line (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
)</li>
<li id="description-tira-cli"><b>TIRA (CLI)</b>: on the command line with <code>tira-run</code> (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
and the
<a href="https://pypi.org/project/tira/">Python TIRA package</a>
)</li>
<li id="description-tira-python"><b>TIRA (Python)</b>: in a Python script (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
and the
<a href="https://pypi.org/project/tira/">Python TIRA package</a>
)</li>
</ul>
"""

        toc = ["## List of software"]
        template_toc_team = "- [Team \u0060$TEAM\u0060](#team-$TEAM_LINK)"
        template_toc_software = "  - [Software \u0060$SOFTWARE\u0060](#software-$SOFTWARE_LINK)"

        content = []
        template_team_header = "\n## Team \u0060$TEAM\u0060\n[See generic instructions above](#instructions)"
        template_software = """### Software \u0060$SOFTWARE\u0060
- [Docker](#description-docker):
  \u0060\u0060\u0060bash
  $CMD_TIRA_DOCKER
  \u0060\u0060\u0060
- [TIRA (CLI)](#description-tira-cli):
  \u0060\u0060\u0060bash
  $CMD_TIRA_CLI
  \u0060\u0060\u0060
- [TIRA (Python)](#description-tira-python):
  \u0060\u0060\u0060python
  $CMD_TIRA_PYTHON
  \u0060\u0060\u0060
"""
        separator = "---"

        softwares = self.all_softwares().sort_values(by=["approach"])

        prev_team = ""
        for _, i in softwares.iterrows():
            execution_info = self.local_execution.run(
                identifier=i["approach"], input_dir="$PWD/input", output_dir="$PWD/output", dry_run=True
            )

            software_name = i["approach"].split("/")[-1]
            software_link = software_name.lower().replace(" ", "-")
            team_name = i["team"]
            team_link = team_name.lower().replace(" ", "-")

            if team_name != prev_team:
                toc.append(template_toc_team.replace("$TEAM_LINK", team_link).replace("$TEAM", team_name))

                if len(prev_team) == 0:
                    pre_text = pre_text.replace("$TASK_ID", "/".join(i["approach"].split("/")[:-2]))
                else:
                    content.append(separator)
                content.append(template_team_header.replace("$TEAM", team_name))
                prev_team = team_name

            toc.append(
                template_toc_software.replace("$SOFTWARE_LINK", software_link).replace("$SOFTWARE", software_name)
            )

            content.append(
                template_software.replace("$SOFTWARE", software_name)
                .replace("$CMD_TIRA_CLI", execution_info["tira-run-cli"])
                .replace("$CMD_TIRA_PYTHON", execution_info["tira-run-python"])
                .replace("$CMD_TIRA_DOCKER", execution_info["docker"])
            )

        print(pre_text)
        print("\n".join(toc))
        print("\n".join(content))

    def __load_evaluators(self):
        evaluators = [json.loads(i) for i in open(self.directory + ".tira/evaluators.jsonl")]
        ret = {i["TIRA_DATASET_ID"]: i for i in evaluators}

        for evaluator in evaluators:
            dataset_id = evaluator["TIRA_DATASET_ID"]
            current_version = version.parse(ret[dataset_id]["TIRA_EVALUATION_IMAGE_TO_EXECUTE"].split(":")[-1])
            available_version = version.parse(evaluator["TIRA_EVALUATION_IMAGE_TO_EXECUTE"].split(":")[-1])

            if available_version > current_version:
                ret[dataset_id] = evaluator

        return ret

    def __load_job_data(self, job_file):
        job = [i.split("=", 1) for i in open(job_file, "r")]
        return {k.strip(): v.strip() for k, v in job}

    def all_evaluators(self) -> pd.DataFrame:
        ret = []
        for i in self.__load_evaluators().values():
            ret += [
                {
                    "dataset": i["TIRA_DATASET_ID"],
                    "image": i["TIRA_EVALUATION_IMAGE_TO_EXECUTE"],
                    "command": i["TIRA_EVALUATION_COMMAND_TO_EXECUTE"],
                }
            ]

        return pd.DataFrame(ret)

    def __extract_image_and_command(self, identifier, evaluator=False):
        softwares = self.___load_softwares() if not evaluator else self.__load_evaluators()

        if identifier in softwares and not evaluator:
            return softwares[identifier]["TIRA_IMAGE_TO_EXECUTE"], softwares[identifier]["TIRA_COMMAND_TO_EXECUTE"]
        if evaluator:
            for k, v in softwares.items():
                if k.startswith(identifier):
                    return (
                        v["TIRA_DATASET_ID"],
                        v["TIRA_EVALUATION_IMAGE_TO_EXECUTE"],
                        v["TIRA_EVALUATION_COMMAND_TO_EXECUTE"],
                    )

        raise ValueError(
            f'There is no {("evaluator" if evaluator else "software")} identified by "{identifier}". Choices are:'
            f" {sorted(list(softwares))}"
        )

    def all_evaluated_appraoches(self) -> pd.DataFrame:
        id_to_software_name = {
            int(i["TIRA_SOFTWARE_ID"].split("docker-software-")[1]): i["TIRA_SOFTWARE_NAME"]
            for i in self.___load_softwares().values()
        }
        ret = []
        for evaluation in glob("*/*/*/evaluation"):
            job_dir = glob(evaluation + "/../job-executed-on*.txt")
            if len(job_dir) != 1:
                raise ValueError("Can not handle multiple job definitions: ", job_dir)

            job_definition = self.__load_job_data(job_dir[0])
            job_identifier = (
                job_definition["TIRA_TASK_ID"]
                + "/"
                + job_definition["TIRA_VM_ID"]
                + "/"
                + id_to_software_name[int(job_definition["TIRA_SOFTWARE_ID"].split("docker-software-")[1])]
            )

            for eval_run in glob(f"{evaluation}/*/output/"):
                try:
                    i = {"approach": job_identifier, "dataset": job_definition["TIRA_DATASET_ID"]}
                    i.update(self.__load_output(eval_run, evaluation=True))
                    ret += [i]
                except Exception:
                    pass

        return pd.DataFrame(ret)

    def docker_software(self, approach, software_id=None):
        ret = []

        if not os.path.exists(self.directory + ".tira/submitted-software.jsonl"):
            from tira.rest_api_client import Client as RestClient

            ret = RestClient().docker_software_details(approach)["context"]
            return {
                "tira_image_name": ret["image"],
                "command": ret["command"],
                "id": None,
                "ids_of_previous_stages": [],
            }

        for _, i in self.all_softwares().iterrows():
            if (approach and i["approach"] == approach) or (
                software_id is not None and str(int(software_id)) == i["id"]
            ):
                ret += [
                    {
                        "tira_image_name": i["image"],
                        "command": i["command"],
                        "id": i["id"],
                        "ids_of_previous_stages": i["ids_of_previous_stages"],
                    }
                ]

        if len(ret) == 1:
            return ret[0]

        raise ValueError(
            f'Could not find a unique software with approach="{approach}" or software_id="{software_id}". Found {ret}'
        )

    def run_was_already_executed_on_dataset(self, approach, dataset):
        return self.get_run_execution_or_none(approach, dataset) is not None

    def get_run_output(self, approach, dataset, allow_without_evaluation=False):
        return self.directory + "/" + self.get_run_execution_or_none(approach, dataset)["run_id"]

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None):
        task, team, software = approach.split("/")
        executions = [json.loads(i) for i in open(self.directory + ".tira/executions.jsonl")]

        executions = [
            i
            for i in executions
            if i["TIRA_TASK_ID"] == task
            and i["TIRA_VM_ID"] == team
            and i["TIRA_SOFTWARE_NAME"] == software
            and i["dataset"] == dataset
        ]

        return (
            None
            if len(executions) < 1
            else {"task": task, "dataset": dataset, "team": team, "run_id": executions[0]["output_directory"]}
        )

    def download_run(self, task, dataset, software, team=None, previous_stage=None, return_metadata=False):
        ret = self.get_run_execution_or_none(f"{task}/{team}/{software}", dataset, previous_stage)
        if not ret:
            raise ValueError(
                f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}",'
                f' software="{software}", team={team}, previous_stage={previous_stage}'
            )

        ret = self.directory + "/" + ret["run_id"]

        if not os.path.exists(ret + "/run.txt"):
            ret = pd.read_json(ret + "/rerank.jsonl.gz", lines=True, orient="columns")
            ret["qid"] = ret["qid"].astype(str)
            ret["docno"] = ret["docno"].astype(str)
            ret["docid"] = ret["docno"]
            ret["query"] = ret["qid"]
        else:
            ret = pd.read_csv(ret + "/run.txt", sep="\\s+", names=["query", "q0", "docid", "rank", "score", "system"])
            ret["query"] = ret["query"].astype(str)
            ret["docid"] = ret["docid"].astype(str)
            ret["docno"] = ret["docid"]
            ret["qid"] = ret["query"]

        if return_metadata:
            return ret, "run_id"
        else:
            return ret



import gzip
import json
import logging
import os
from glob import glob
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Union

import pandas as pd


def parse_jsonl_line(input: Union[str, bytearray, bytes], load_default_text: bool) -> Dict:
    """
    Deseralizes the line using JSON deserialization. Optionally strips the 'original_query' and 'original_document'
    fields from the resulting object and converts the qid and docno fields to strings.

    :param str | bytearray | bytes input: A json-serialized string.
    :param bool load_default_text: If true, the original_query and original_document fields are removed and the qid and
        docno values are converted to strings.
    :return: The deserialized and (optionally) processed object.
    :rtype: dict

    :Example:
        >>> parse_jsonl_line('{}', False)
        {}
        >>> parse_jsonl_line('{"original_query": "xxxx"}', False)
        {'original_query': 'xxxx'}
        >>> parse_jsonl_line('{"original_query": "xxxx"}', True)
        {}
        >>> parse_jsonl_line('{"original_query": "xxxx", "qid": 42, "pi": 3.14}', False)
        {'original_query': 'xxxx', 'qid': 42, 'pi': 3.14}
        >>> parse_jsonl_line('{"original_query": "xxxx", "qid": 42, "pi": 3.14}', True)
        {'qid': '42', 'pi': 3.14}
    """
    obj = json.loads(input)
    assert isinstance(obj, dict)
    if load_default_text:
        for field_to_del in ["original_query", "original_document"]:
            obj.pop(field_to_del, None)

        for field_to_str in ["qid", "docno"]:
            if field_to_str in obj:
                obj[field_to_str] = str(obj[field_to_str])

    return obj


def stream_all_lines(input_file: Union[str, Iterable[bytes]], load_default_text: bool) -> Generator[Dict, Any, Any]:
    """
    .. todo:: add documentation
    .. todo:: this function has two semantics: handling a file and handling file-contents
    """
    if type(input_file) is str:
        if not os.path.isfile(input_file):
            return

        if input_file.endswith(".gz"):
            with gzip.open(input_file, "rt", encoding="utf-8") as f:
                yield from stream_all_lines(f, load_default_text)
        else:
            with open(input_file, "r") as f:
                yield from stream_all_lines(f, load_default_text)

        return

    for line in input_file:
        yield parse_jsonl_line(line, load_default_text)


def huggingface_model_mounts(models: Iterable[str]) -> dict:
    """Determine the mounts to make the described huggingface models available in the container. The models must
    already exist in the local huggingface cache of the host.

    Args:
        models (Iterable[str]): A list of huggingface models that you want to mount, e.g., ["openai-community/gpt2"].

    Returns:
        dict: The mounts required to make the specified models available in the container.
    """
    hf_cache = Path(os.path.expanduser("~/.cache/huggingface/hub"))
    if "HF_HUB_CACHE" in os.environ:
        hf_cache = Path(os.environ["HF_HUB_CACHE"])
    elif "HF_HOME" in os.environ:
        hf_cache = Path(os.environ["HF_HOME"]) / "hub"
    ret = {}
    if not models:
        return ret

    hf_cache = hf_cache.absolute()
    for model in models:
        model_in_fs = ("models/" + str(model)).replace("/", "--")
        model_path = hf_cache / model_in_fs
        if not model_path.exists():
            msg = (
                f"Model {model} does not in HF_HUB_CACHE = '{hf_cache}'. Expected a directory '{model_path}' to exist."
                " Please ensure that the model is available via your preferred way to download it."
            )
            print(msg)
            raise ValueError(msg)
        else:
            ret[str(model_path)] = {"bind": f"/root/.cache/huggingface/hub/{model_in_fs}", "mode": "ro"}

    return ret


def change_workdir_cmd(workdir: str):
    """
    Returns a shell command that changes the working directory to the specified directory to be executed in the TIRA
    sandbox before the real command.
    """

    if not workdir:
        return 'echo "did not change the working directory"'

    workdir = workdir.strip()

    if not workdir or not workdir.startswith("/"):
        return 'echo "did not change the working directory"'

    return f'cd {workdir}; echo "changed workdir to {workdir}"'


def _ln_huggingface_model_mounts(models: str) -> str:
    """Create a set of ln statements for symbolic links to huggingface models within a running container. Fails if the
    models do not exist in the local huggingface cache of the host.

    Args:
        models (str): A space seperated list of huggingface models that you want to mount, e.g.,
            "openai-community/gpt2 openai-community/gpt2-large".

    Returns:
        str: the ln command for usage with eval.
    """
    if not models:
        return ""
    models = sorted(list(set([i for i in models.split(" ") if i])))

    if not models:
        return ""

    ret = ["mkdir -p /root/.cache/huggingface/hub/"]
    for model in models:
        i = huggingface_model_mounts([model])
        k = list(i.keys())
        if not k or not len(k) == 1:
            raise ValueError(f"Expected exactly one model to be mounted, got: {i}")
        target_dir = i[k[0]]["bind"]
        src_dir = k[0]

        ret += [f"rm -Rf {target_dir}", f"ln -s {src_dir} {target_dir}"]

    return "; ".join(ret + [f'echo "mounted {len(models)} models"'])


def all_lines_to_pandas(input_file: Union[str, Iterable[str]], load_default_text: bool) -> pd.DataFrame:
    """
    .. todo:: add documentation
    .. todo:: this function has two semantics: handling a file and handling file-contents
    """
    if type(input_file) is str:
        if input_file.endswith(".gz"):
            with gzip.open(input_file, "rt", encoding="utf-8") as f:
                return all_lines_to_pandas(f, load_default_text)
        else:
            with open(input_file, "r") as f:
                return all_lines_to_pandas(f, load_default_text)

    import pandas as pd

    ret = []

    for line in input_file:
        ret += [parse_jsonl_line(line, load_default_text)]

    return pd.DataFrame(ret)


def __num(input: str) -> Union[str, int, float]:
    """
    Converts the input to an int or float if possible. Returns the inputted string otherwise.

    :param str input: The string that should be converted to a float or int if possible.
    :return: The intrepteted input.
    :rtype: str | int | float

    :Example:
        >>> __num("hello world")
        "hello world"
        >>> __num("-42")
        -42
        >>> __num("3.5")
        3.5
        >>> __num("2e-6")
        2e-6
        >>> __num(" -42")
        " -42"
    """
    try:
        return int(input)
    except ValueError:
        try:
            return float(input)
        except ValueError:
            return input


def run_cmd(cmd: List[str], ignore_failure=False):
    import subprocess

    exit_code = subprocess.call(cmd)

    if not ignore_failure and exit_code != 0:
        raise ValueError(f"Command {cmd} did exit with return code {exit_code}.")


def parse_prototext_key_values(file_name):
    for i in [i for i in open(file_name, "r").read().split("measure {")]:
        ret = {}
        for line in i.split("\n"):
            if len(line.split(":")) < 2:
                continue
            elif len(line.split(":")) > 2:
                raise ValueError(f'Could not parse "{line}".')
            key = line.split(":")[0]
            value = line.split(":")[1].split('"')[1]
            ret[key.strip()] = __num(value.strip())
        if len(ret) > 0:
            yield ret


def to_prototext(m: List[Dict[str, Any]], upper_k: str = "") -> str:
    ret = ""

    def _to_prototext(d: Dict[str, Any], upper_k: str = "") -> str:
        ret = ""
        for k, v in d.items():
            new_k = upper_k
            if not new_k:
                new_k = k
            elif not new_k.endswith(k):
                new_k = upper_k + "_" + k
            if isinstance(v, dict):
                ret += _to_prototext(v, upper_k=new_k)
            else:
                ret += 'measure{\n  key: "' + str(new_k.replace("_", " ").title()) + '"\n  value: "' + str(v) + '"\n}\n'
        return ret

    for d in m:
        ret += _to_prototext(d, upper_k=upper_k)

    return ret


def all_environment_variables_for_github_action_or_fail(params):
    ret = {}

    for i in params:
        if len(i.split("=")) != 2:
            raise ValueError(f"Expect that exactly one '=' occurs in each passed parameter, got: '{i}'")

        key, value = i.split("=")

        if key in ret:
            raise ValueError(f"Got duplicated key: '{key}'")

        ret[key] = value

    expected_keys = [
        "GITHUB_SHA",
        "TIRA_VM_ID",
        "TIRA_TASK_ID",
        "TIRA_DOCKER_REGISTRY_TOKEN",
        "TIRA_DOCKER_REGISTRY_USER",
        "TIRA_CLIENT_TOKEN",
        "TIRA_CLIENT_USER",
        "TIRA_CODE_REPOSITORY_ID",
    ]

    for k in expected_keys:
        if k not in ret or not ret[k]:
            raise ValueError(
                f"I need the parameter {k} to continue, but could not find one or it is empty. This likely is a"
                " configuration error, e.g., due to missing secrets."
            )

    if "TIRA_JUPYTER_NOTEBOOK" in ret:
        for to_del in ["TIRA_DOCKER_FILE", "TIRA_DOCKER_PATH", "TIRA_COMMAND"]:
            if to_del in ret:
                del ret[to_del]
        ret["IMAGE_TAG"] = (
            f'registry.webis.de/code-research/tira/tira-user-{ret["TIRA_VM_ID"]}/submission:{ret["GITHUB_SHA"]}'
        )
    else:
        for expected_key in ["TIRA_DOCKER_FILE", "TIRA_DOCKER_PATH", "TIRA_COMMAND"]:
            if k not in ret or not ret[k]:
                raise ValueError(
                    f"I need the parameter {k} to continue, but could not find one or it is empty. This likely is a"
                    " configuration error, e.g., due to missing secrets."
                )

            submission_id = f'{ret["TIRA_DOCKER_PATH"].replace("/", "-").replace(" ", "-")}:{ret["GITHUB_SHA"]}'
            ret["IMAGE_TAG"] = (
                f'registry.webis.de/code-research/tira/tira-user-{ret["TIRA_VM_ID"]}/submission-{submission_id}'
            )

    return [k + "=" + v for k, v in ret.items()]


def load_output_of_directory(directory: Path, evaluation: bool = False) -> Union[Dict, pd.DataFrame]:
    files = glob(str(directory) + "/*")

    if evaluation:
        files = [i for i in files if i.endswith(".prototext")]

    if len(files) != 1:
        raise ValueError("Expected exactly one output file. Got: ", files)

    files = files[0]

    logging.debug(f"Read file from {files}")

    if evaluation:
        ret = {}
        for i in [i for i in open(files, "r").read().split("measure") if "key:" in i and "value:" in i]:
            key = i.split("key:")[1].split("value")[0].split('"')[1]
            value = i.split("key:")[1].split("value")[1].split('"')[1]

            ret[key.strip()] = __num(value.strip())

        return ret
    else:
        return pd.read_json(files, lines=True, orient="records")



import hashlib
import io
import json
import logging
import os
import time
import zipfile
from functools import lru_cache
from glob import glob
from pathlib import Path
from random import randint
from typing import Dict, List, Optional, Union

import pandas as pd
import requests
from tqdm import tqdm

from tira.check_format import _fmt, check_format
from tira.local_execution_integration import LocalExecutionIntegration
from tira.pandas_integration import PandasIntegration
from tira.profiling_integration import ProfilingIntegration
from tira.pyterrier_integration import PyTerrierAnceIntegration, PyTerrierIntegration, PyTerrierSpladeIntegration
from tira.third_party_integrations import temporary_directory
from tira.tira_redirects import (
    RESOURCE_REDIRECTS,
    TASKS_WITH_REDIRECT_MERGING,
    dataset_ir_redirects,
    mirror_url,
    redirects,
)

from .tira_client import TiraClient


class Client(TiraClient):
    base_url: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        api_key: str = None,
        failsave_retries: int = 5,
        failsave_max_delay: int = 15,
        api_user_name: Optional[str] = None,
        tira_cache_dir: Optional[str] = None,
        verify: bool = True,
    ):
        self.base_url = base_url or "https://www.tira.io"
        self.tira_cache_dir = (
            tira_cache_dir if tira_cache_dir else os.environ.get("TIRA_CACHE_DIR", os.path.expanduser("~") + "/.tira")
        )
        self.json_cache = {}

        if api_key is None:
            self.api_key = self.load_settings()["api_key"]
            self.api_user_name = self.load_settings()["api_user_name"]
        else:
            self.api_key = api_key
            self.api_user_name = api_user_name

        self.failsave_retries = 1
        if self.api_key != "no-api-key":
            self.fail_if_api_key_is_invalid()
        self.pd = PandasIntegration(self)
        self.pt = PyTerrierIntegration(self)
        self.profiling = ProfilingIntegration(self)
        self.pt_ance = PyTerrierAnceIntegration(self)
        self.pt_splade = PyTerrierSpladeIntegration(self)
        self.local_execution = LocalExecutionIntegration(self)

        self.failsave_retries = failsave_retries
        self.failsave_max_delay = failsave_max_delay
        self.verify = verify

    def load_settings(self):
        try:
            ret = json.load(open(self.tira_cache_dir + "/.tira-settings.json", "r"))
            if "api_key" not in ret:
                ret["api_key"] = "no-api-key"
            if "api_user_name" not in ret:
                ret["api_user_name"] = "no-api-key-user"
            return ret
        except Exception:
            logging.info(f"No settings given in {self.tira_cache_dir}/.tira-settings.json. I will use defaults.")
            return {"api_key": "no-api-key", "api_user_name": "no-api-key-user"}

    def update_settings(self, k, v):
        settings = self.load_settings()
        settings[k] = v
        os.makedirs(self.tira_cache_dir, exist_ok=True)
        json.dump(settings, open(Path(self.tira_cache_dir) / ".tira-settings.json", "w+"))

        if k == "api_key":
            self.api_key = settings["api_key"]

    def api_key_is_valid(self):
        role = self.json_response("/api/role")

        if (
            (self.api_user_name is None or self.api_user_name == "no-api-key-user")
            and role
            and "context" in role
            and "user_id" in role["context"]
            and role["context"]["user_id"]
        ):
            self.api_user_name = role["context"]["user_id"]

        return (
            role
            and "status" in role
            and "role" in role
            and "csrf" in role
            and role["csrf"]
            and 0 == role["status"]
            and "user_id" in role["context"]
            and role["context"]["user_id"]
            and "role" in role["context"]
            and role["context"]["role"]
            and "guest" != role["context"]["role"]
        )

    def fail_if_api_key_is_invalid(self):
        if not self.api_key_is_valid():
            role = self.json_response("/api/role")
            raise ValueError("It seems like the api key is invalid. Got: ", role)

    def datasets(self, task):
        return json.loads(self.json_response(f"/api/datasets_by_task/{task}")["context"]["datasets"])

    def get_dataset(self, dataset) -> dict:
        """Get the TIRA representation of an dataset identified by the passed dataset argument.

        Args:
            dataset (Union[str, IrDataset, PyTerrierDataset): The dataset that is either the string id of the dataset in TIRA, the string id of an ir_dataset, the string id of an PyTerrier dataset, or an instantiation of an ir_dataset or an PyTerrier dataset.
        Returns:
            dict: The TIRA representation of the dataset.
        """

        dataset_identifier = self._TiraClient__extract_dataset_identifier(dataset)
        datasets = self.archived_json_response("/v1/datasets/all")

        ret = self._TiraClient__matching_dataset(datasets, dataset_identifier)
        if ret is not None:
            return ret

        # retry with force_reload.
        datasets = self.archived_json_response("/v1/datasets/all", force_reload=True)
        ret = self._TiraClient__matching_dataset(datasets, dataset_identifier)
        if ret is not None:
            return ret

        msg = f'The dataset "{dataset_identifier}" is not publicly available in TIRA. Please visit https://tira.io/datasets for an overview of all public datasets.'
        print(msg)
        raise ValueError(msg)

    def docker_software_id(self, approach):
        return self.docker_software(approach)["docker_software_id"]

    def all_softwares(self, task_id):
        """
        Return all public submissions.
        """
        return [
            task_id + "/" + i["vm_id"] + "/" + i["display_name"]
            for i in self.json_response(f"/api/task/{task_id}/public-submissions")["context"]["public_submissions"]
        ]

    def all_tasks(self):
        return self.json_response(f"/tira-backend/api/task-list")["context"]["task_list"]

    def docker_software(self, approach):
        task, team, software = approach.split("/")
        return self.json_response(f"/api/task/{task}/submission-details/{team}/{software}")["context"]["submission"]

    def docker_credentials(self, task_name, team_name):
        ret = self.metadata_for_task(task_name, team_name)
        if ret and "status" in ret and ret["status"] == 0 and "context" in ret and "docker" in ret["context"]:
            return (
                ret["context"]["docker"]["docker_registry_user"],
                ret["context"]["docker"]["docker_registry_token"],
                self.docker_registry(),
            )
        return None, None, self.docker_registry()

    def docker_software_details(self, approach):
        task, team, software = approach.split("/")
        ret = self.json_response(f"/task/{task}/vm/{team}/software_details/{software}")

        return ret

    def metadata_for_task(self, task_name, team_name=None):
        if team_name is None:
            return self.json_response(f"/api/task/{task_name}")
        else:
            return self.json_response(f"/api/task/{task_name}/user/{team_name}")

    def add_docker_software(
        self,
        image,
        command,
        tira_vm_id,
        tira_task_id,
        code_repository_id,
        build_environment,
        previous_stages=[],
        mount_hf_model=[],
    ):
        headers = {
            "Api-Key": self.api_key,
            "Api-Username": self.api_user_name,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.fail_if_api_key_is_invalid()
        url = f"{self.base_url}/task/{tira_task_id}/vm/{tira_vm_id}/add_software/docker"
        content = {
            "action": "post",
            "image": image,
            "command": command,
            "code_repository_id": code_repository_id,
            "build_environment": json.dumps(build_environment),
        }

        if previous_stages and len(previous_stages) > 0:
            content["inputJob"] = previous_stages

        if mount_hf_model and len(mount_hf_model) > 0:
            content["mount_hf_model"] = mount_hf_model

        ret = requests.post(url, headers=headers, json=content)
        response_code = ret.status_code
        ret = ret.content.decode("utf8")
        try:
            ret = json.loads(ret)
        except Exception:
            msg = f"Upload of software failed with error {ret} and response code {response_code}."
            print(msg)
            raise ValueError(msg)

        if ret["status"] != 0:
            msg = f"Upload of software failed with error {ret}"
            print(msg)
            raise ValueError(msg)

        print(f'Software with name {ret["context"]["display_name"]} was created.')
        logging.info(f'Software with name {ret["context"]["display_name"]} was created.')
        logging.info(
            f"Please visit {self.base_url}/submit/{tira_task_id}/user/{tira_vm_id}/docker-submission to run your"
            " software."
        )

    def submissions(self, task, dataset):
        response = self.json_response(f"/api/submissions/{task}/{dataset}")["context"]
        ret = []

        for vm in response["vms"]:
            for run in vm["runs"]:
                if "review" in run:
                    for k, v in run["review"].items():
                        run["review_" + k] = v
                    del run["review"]

                ret += [
                    {**{"task": response["task_id"], "dataset": response["dataset_id"], "team": vm["vm_id"]}, **run}
                ]

        return pd.DataFrame(ret)

    def upload_submissions(self, task_id, vm_id, upload_id, dataset=None):
        ret = self.json_response(f"/api/upload-group-details/{task_id}/{vm_id}/{upload_id}")
        ret = ret["context"]["upload_group_details"]["runs"]

        return [i for i in ret if dataset is None or dataset == i["dataset"]]

    def submissions_with_evaluation_or_none(self, task, dataset, team, software):
        """This method returns all runs of the specified software in the task on the dataset by the team.
        This is especially suitable to batch evaluate all submissions of the software because the evaluation is none if
        no successfull evaluation was conducted (or there is a new evaluator).
        E.g., by code like:

        .. code:: py

            for approach in ['approach-1', ..., 'approach-n]:
                runs_for_approach = tira.submissions_with_evaluation_or_none(task, dataset, team, approach)
                for i in runs_for_approach:
                    if not i['evaluation']:
                        tira.evaluate_run(team, dataset, i['run_id'])
        """
        submissions = self.submissions(task, dataset)
        evaluations = self.evaluations(task, dataset, join_submissions=False)
        run_to_evaluation = {}
        for _, i in evaluations.iterrows():
            i = i.to_dict()
            run_id = i["run_id"]
            del i["task"]
            del i["dataset"]
            del i["team"]
            del i["run_id"]
            run_to_evaluation[run_id] = i

        if len(submissions) < 1:
            return []

        submissions = submissions[
            (submissions["task"] == task)
            & (submissions["dataset"] == dataset)
            & (submissions["team"] == team)
            & (submissions["software"] == software)
        ]
        ret = []
        for run_id in submissions.run_id.unique():
            ret += [
                {
                    "run_id": run_id,
                    "task": task,
                    "dataset": dataset,
                    "team": team,
                    "software": software,
                    "evaluation": run_to_evaluation.get(run_id, None),
                }
            ]

        return ret

    def evaluations(self, task, dataset, join_submissions=True):
        response = self.json_response(f"/api/evaluations/{task}/{dataset}")["context"]
        ret = []
        evaluation_keys = response["ev_keys"]

        if join_submissions:
            runs_to_join = {}
            for _, i in self.submissions(task, dataset).iterrows():
                i = i.to_dict()
                runs_to_join[(i["team"], i["run_id"])] = {
                    "software": i["software"],
                    "input_run_id": i["input_run_id"],
                    "is_upload": i["is_upload"],
                    "is_docker": i["is_docker"],
                }

        for evaluation in response["evaluations"]:
            run = {
                "task": response["task_id"],
                "dataset": response["dataset_id"],
                "team": evaluation["vm_id"],
                "run_id": evaluation["input_run_id"],
                "evaluation_run_id": evaluation["run_id"],
                "published": evaluation["published"],
                "blinded": evaluation["blinded"],
            }

            if join_submissions and (run["team"], run["run_id"]) in runs_to_join:
                software = runs_to_join[(run["team"], run["run_id"])]
                for k, v in software.items():
                    run[k] = v

            for measure_id, measure in zip(range(len(evaluation_keys)), evaluation_keys):
                run[measure] = evaluation["measures"][measure_id]

            ret += [run]

        return pd.DataFrame(ret)

    def run_was_already_executed_on_dataset(self, approach, dataset):
        return self.get_run_execution_or_none(approach, dataset) is not None

    def load_resource(self, resource: str):
        """Load a resource (usually a zip) from TIRA/Zenodo. Serves as utikity function in case some additional
        resources must be loaded.

        Args:
            resource (str): The resource identifier

        Raises:
            ValueError: If the resource is not known.

        Returns:
            str: The path to the downloaded resource.
        """
        target_file = f"{self.tira_cache_dir}/raw_resources/{resource}"
        if os.path.exists(target_file):
            return target_file

        if resource not in RESOURCE_REDIRECTS:
            raise ValueError(f"Resource {resource} not supported.")

        os.makedirs(f"{self.tira_cache_dir}/raw_resources", exist_ok=True)
        self.download_and_extract_zip(RESOURCE_REDIRECTS[resource], target_file, extract=False)
        return target_file

    def get_run_output(self, approach, dataset, allow_without_evaluation=False):
        """
        Downloads the run (or uses the cached version) of the specified approach on the specified dataset.
        Returns the directory containing the outputs of the run.
        """
        mounted_output_in_sandbox = self.input_run_in_sandbox(approach)
        if mounted_output_in_sandbox:
            return mounted_output_in_sandbox

        task, team, software = approach.split("/")
        if "/" in dataset:
            dataset = dataset.split("/")[-1]

        run_execution = self.get_run_execution_or_none(approach, dataset)

        if run_execution:
            return self.download_zip_to_cache_directory(task, dataset, team, run_execution["run_id"])

        run_execution = self.submissions_with_evaluation_or_none(task, dataset, team, software)
        run_execution = [
            i for i in run_execution if ("evaluation" in i and i["evaluation"]) or allow_without_evaluation
        ]

        if run_execution is None or len(run_execution) < 1:
            raise ValueError(f'Could not get run for approach "{approach}" on dataset "{dataset}".')

        return self.download_zip_to_cache_directory(
            run_execution[0]["task"], run_execution[0]["dataset"], run_execution[0]["team"], run_execution[0]["run_id"]
        )

    def public_runs(self, task, dataset, team, software):
        ret = self.json_response(f"/api/list-runs/{task}/{dataset}/{team}/" + software.replace(" ", "%20"))
        if ret and "context" in ret and "runs" in ret["context"] and ret["context"]["runs"]:
            return ret["context"]
        else:
            return None

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None):
        task, team, software = approach.split("/")
        redirect = redirects(approach, dataset)

        if redirect is not None and "run_id" in redirect and redirect["run_id"] is not None:
            return {"task": task, "dataset": dataset, "team": team, "run_id": redirect["run_id"]}

        public_runs = self.public_runs(task, dataset, team, software)
        if public_runs:
            return {"task": task, "dataset": dataset, "team": team, "run_id": public_runs["runs"][0]}

        if not self.api_key_is_valid():
            raise ValueError(
                f'No public submissions for "{approach} on "{dataset}" and you are not authenticated. Please authenticate to access private submissions'
            )

        df_eval = self.submissions_of_team(task=task, dataset=dataset, team=team)
        if len(df_eval) <= 0:
            return None

        ret = df_eval[(df_eval["dataset"] == dataset) & (df_eval["software"] == software)]
        if len(ret) <= 0:
            return None

        if team:
            ret = ret[ret["team"] == team]

        # FIXME: Is this really necessary or is it checked with the if len(ret) <= 0 later on?
        if len(ret) <= 0:
            return None

        if previous_stage_run_id:
            ret = ret[ret["input_run_id"] == previous_stage_run_id]

        if len(ret) <= 0:
            return None

        return ret[["task", "dataset", "team", "run_id"]].iloc[0].to_dict()

    def download_run(self, task, dataset, software, team=None, previous_stage=None, return_metadata=False):
        mounted_output_in_sandbox = self.input_run_in_sandbox(f"{task}/{team}/{software}")
        if mounted_output_in_sandbox:
            ret = pd.read_csv(
                mounted_output_in_sandbox + "/run.txt",
                sep="\\s+",
                names=["query", "q0", "docid", "rank", "score", "system"],
                dtype={"query": str, "docid": str},
            )
            if return_metadata:
                return ret, "run-id"
            else:
                return ret

        if "/" in dataset:
            dataset = dataset.split("/")[-1]
        ret = self.get_run_execution_or_none(f"{task}/{team}/{software}", dataset, previous_stage)
        if not ret:
            raise ValueError(
                f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}",'
                f' software="{software}", team={team}, previous_stage={previous_stage}'
            )
        run_id = ret["run_id"]

        ret = self.download_zip_to_cache_directory(**ret)
        ret = pd.read_csv(
            ret + "/run.txt",
            sep="\\s+",
            names=["query", "q0", "docid", "rank", "score", "system"],
            dtype={"query": str, "docid": str},
        )
        if return_metadata:
            return ret, run_id
        else:
            return ret

    def download_evaluation(self, task, dataset, software, team):
        ret = self.submissions_with_evaluation_or_none(task, dataset, team, software)
        if not ret or len(ret) < 1:
            raise ValueError(
                f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}",'
                f' software="{software}", team={team}.'
            )
        run_id = ret[0]["run_id"]

        submissions = self.submissions(task, dataset)
        submissions = submissions[(submissions["input_run_id"] == run_id) & (submissions["is_evaluation"])]

        if submissions is None or len(submissions) < 1:
            raise ValueError(
                f'I could not find a evaluation for the filter criteria task="{task}", dataset="{dataset}",'
                f' software="{software}", team={team}, run_id={run_id}.'
            )

        return self.download_zip_to_cache_directory(task, dataset, team, submissions.iloc[0].to_dict()["run_id"])

    def download_dataset(self, task, dataset, truth_dataset=False):
        """
        Download the dataset. Set truth_dataset to true to load the truth used for evaluations.
        """
        if "TIRA_INPUT_DATASET" in os.environ:
            return os.environ["TIRA_INPUT_DATASET"]
        if "/" in dataset:
            dataset = dataset.split("/")[-1]

        meta_data = self.get_dataset(f"{task}/{dataset}")
        data_type = "training" if dataset.endswith("-training") else "test"
        suffix = "inputs" if not truth_dataset else "truths"
        url = None
        if (
            not meta_data
            or "mirrors" not in meta_data
            or suffix not in meta_data["mirrors"]
            or not meta_data["mirrors"][suffix]
        ):
            dataset = dataset_ir_redirects(dataset)
        else:
            url = list(meta_data["mirrors"][suffix].values())[0]

        target_dir = f"{self.tira_cache_dir}/extracted_datasets/{task}/{dataset}/"
        suffix = "input-data" if not truth_dataset else "truth-data"
        if os.path.isdir(target_dir + suffix):
            return target_dir + suffix

        if not url:
            url = f'{self.base_url}/data-download/{data_type}/input-{("" if not truth_dataset else "truth")}/{dataset}.zip'

        self.download_and_extract_zip(url, target_dir)

        os.rename(target_dir + f"/{dataset}", target_dir + suffix)

        return target_dir + suffix

    def download_zip_to_cache_directory(self, task, dataset, team, run_id):
        target_dir = f"{self.tira_cache_dir}/extracted_runs/{task}/{dataset}/{team}"

        if os.path.isdir(target_dir + f"/{run_id}"):
            return target_dir + f"/{run_id}/output"

        potential_local_matches = glob(f"{self.tira_cache_dir}/extracted_runs/{task}/{dataset}/*/{run_id}/output")
        if task in TASKS_WITH_REDIRECT_MERGING and len(potential_local_matches) == 1:
            return potential_local_matches[0]

        self.download_and_extract_zip(
            f"{self.base_url}/task/{task}/user/{team}/dataset/{dataset}/download/{run_id}.zip", target_dir
        )

        return target_dir + f"/{run_id}/output"

    def add_run_to_leaderboard(self, task, team, dataset, evaluation_run_id=None, run_id=None):
        """
        Publish the specified run to the leaderboard.

        This is especially suitable to batch add all submissions of submissions, e.g., by code like:

        .. code:: py

            for approach in ['approach-1', ..., 'approach-n']:
                runs_for_approach = tira.submissions_with_evaluation_or_none(task, dataset, team, approach)
                for i in runs_for_approach:
                    if i['evaluation']:
                        tira.add_run_to_leaderboard(TASK, team, dataset, run_id=i['run_id'])

        """
        if run_id and evaluation_run_id:
            raise ValueError(
                "Please pass either a evaluation_run_id or a run_id, but both were passed:"
                f" evaluation_run_id={evaluation_run_id}, run_id={run_id}"
            )
        if run_id and not evaluation_run_id:
            submissions = self.submissions(task, dataset)
            submissions = submissions[(submissions["input_run_id"] == run_id) & (submissions["is_evaluation"])]

            for evaluation_run_id in submissions["run_id"].unique():
                self.add_run_to_leaderboard(task, team, dataset, evaluation_run_id=evaluation_run_id)

        if evaluation_run_id:
            logging.info(f"Publish run: {evaluation_run_id}.")
            ret = self.json_response(f"/publish/{team}/{dataset}/{evaluation_run_id}/true")

            if ("status" not in ret) or ("0" != ret["status"]) or ("published" not in ret) or (not ret["published"]):
                raise ValueError(f"Adding the run to the leaderboard failed. Got {ret}")

    def get_configuration_of_evaluation(self, task_id, dataset_id):
        """Get the configuration of the evaluator for the passed dataset inside the task specified by task_id."""
        ret = self.json_response(f"/api/configuration-of-evaluation/{task_id}/{dataset_id}")

        if "status" not in ret or "0" != str(ret["status"]):
            raise ValueError(f"Failed to load configuration of an evaluator. Got {ret}")

        return ret["context"]["dataset"]

    def evaluate_run(self, team, dataset, run_id):
        """Evaluate the run of the specified team and identified by the run_id (the run must be submitted on the
        specified dataset)."""
        ret = self.json_response(f"/grpc/{team}/run_eval/{dataset}/{run_id}")

        if "status" not in ret or "0" != str(ret["status"]):
            raise ValueError(f"Failed to evaluate the run. Got {ret}")

        return ret

    def download_and_extract_zip(self, url, target_dir, extract=True):
        url = redirects(url=url)["urls"][0]
        if url.split("://")[1].startswith("files.webis.de"):
            print(f"Download from the Incubator: {url}")
            print("\tThis is only used for last spot checks before archival to Zenodo.")

        if url.split("://")[1].startswith("zenodo.org"):
            print(f"Download from Zenodo: {url}")

        for _ in range(self.failsave_retries):
            status_code = None
            try:
                headers = {"Api-Key": self.api_key, "Api-Username": self.api_user_name}
                if self.api_key == "no-api-key":
                    del headers["Api-Key"]
                if self.api_user_name == "no-api-key-user":
                    del headers["Api-Username"]

                r = requests.get(url, headers=headers, stream=True)
                total = int(r.headers.get("content-length", 0))
                status_code = r.status_code
                if status_code < 200 or status_code >= 300:
                    raise ValueError(f"Got non 200 status code {status_code} for {url}.")

                response_content = io.BytesIO()
                with tqdm(
                    desc="Download",
                    total=total,
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar:
                    for data in r.iter_content(chunk_size=1024):
                        size = response_content.write(data)
                        bar.update(size)
                if extract:
                    print("Download finished. Extract...")
                    z = zipfile.ZipFile(response_content)
                    z.extractall(target_dir)
                    print("Extraction finished: ", target_dir)
                else:
                    print("Download finished. Persist...")
                    with open(target_dir, "wb") as file_out:
                        file_out.write(response_content.getbuffer())
                    print("Download finished: ", target_dir)

                return
            except Exception:
                sleep_time = randint(1, self.failsave_max_delay)
                print(f"Code: {status_code}")
                print(f"Error occured while fetching {url}. I will sleep {sleep_time} seconds and continue.")
                url = mirror_url(url)
                time.sleep(sleep_time)

    def login(self, token):
        self.api_key = token

        if not self.api_key_is_valid():
            print("The api key {token} is not valid")
            raise ValueError(f"The api key {token} is invalid.")

        self.update_settings("api_key", token)

    def get_authentication_cookie(self, user, password):
        resp = requests.get(f"{self.base_url}/session/csrf", headers={"x-requested-with": "XMLHttpRequest"})

        header = {
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            "cookie": "_forum_session=" + resp.cookies["_forum_session"],
            "x-csrf-token": resp.json()["csrf"],
            "x-requested-with": "XMLHttpRequest",
        }

        resp = requests.post(f"{self.base_url}/session", data=f"login={user}&password={password}", headers=header)

        return f'_t={resp.cookies["_t"]}; _forum_session={resp.cookies["_forum_session"]}'

    def run_software(self, approach, dataset, resources, rerank_dataset="none", software_id=None):
        task, team, software = approach.split("/")
        authentication_cookie = self.get_authentication_cookie(
            self.load_settings()["user"], self.load_settings()["password"]
        )

        if not software_id:
            software_id = self.docker_software_id(approach)
        if not software_id:
            raise ValueError(f'Could not find software id for "{approach}". Got: "{software_id}".')

        url = (
            f"{self.base_url}/grpc/{task}/{team}/run_execute/docker/"
            "{dataset}/{software_id}/{resources}/{rerank_dataset}"
        )
        logging.info(f"Start software...\n\t{url}\n")

        csrf_token = self.get_csrf_token()
        headers = {
            # 'Api-Key': self.api_key,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Cookie": authentication_cookie,
            "x-csrftoken": csrf_token,
        }

        ret = requests.post(url, headers=headers, json={"csrfmiddlewaretoken": csrf_token, "action": "post"})
        ret = ret.content.decode("utf8")
        logging.info(ret)
        ret = json.loads(ret)
        assert ret["status"] == 0

    def get_upload_group_id(self, task_id: str, vm_id: str, display_name: str) -> int:
        """Get the id of the upload group of user specified with vm_id for the task task_id with the display_name.
        Raises an error if no matching upload_group was found."""
        url = f"/api/submissions-for-task/{task_id}/{vm_id}/upload"
        ret = self.json_response(url)
        if "context" not in ret or "all_uploadgroups" not in ret["context"]:
            logging.error("Failed to get upload id, response does not contain the expected fields.")
            raise ValueError(f"Invalid response for request {url}: {ret}.")

        for upload_group in ret["context"]["all_uploadgroups"]:
            if display_name is not None and upload_group["display_name"] == display_name:
                return upload_group["id"]

        logging.error(
            f"Could not find upload with display_name {display_name} for task {task_id} of user {vm_id}. Got:",
            ret["context"]["all_uploadgroups"],
        )
        raise ValueError(
            "Could not find upload with display_name {display_name} for task {task_id} of user {vm_id}. Got:",
            ret["context"]["all_uploadgroups"],
        )

    def create_upload_group(self, task_id: str, vm_id: str, display_name: str) -> Optional[str]:
        # TODO: check that task_id and vm_id don't contain illegal characters (e.g., '/')
        # TODO: Make this idempotent: reuse existing upload group if it already exists.
        url = f"{self.base_url}/task/{task_id}/vm/{vm_id}/add_software/upload"
        logging.debug(f"Creating a new upload at {url}")
        ret = self.json_response(url)

        logging.debug(f"Created new upload with id {ret['upload']}")
        return ret["upload"]

    def upload_run_anonymous(self, file_path: Path, dataset_id: str):
        upload_to_tira = self.get_dataset(dataset_id)

        if isinstance(file_path, str):
            file_path = Path(file_path)

        # TODO use format from upload_to_tira instead of hard-coded run.txt
        status_code, msg = check_format(file_path, "run.txt")

        if status_code != _fmt.OK:
            print(msg)
            raise ValueError(msg)

        zip_file = temporary_directory()
        zip_file = zip_file / "tira-upload.zip"

        zf = zipfile.ZipFile(zip_file, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=9)
        for root, _, files in os.walk(file_path):
            for name in files:
                filePath = os.path.join(root, name)
                zf.write(filePath, arcname=name)

        zf.close()
        headers = {"Accept": "application/json"}
        files = {"file": open(zip_file, "rb")}

        resp = requests.post(
            url=f"{self.base_url}/api/v1/anonymous-uploads/{upload_to_tira['dataset_id']}",
            files=files,
            headers=headers,
            verify=self.verify,
        )

        if resp.status_code not in {200, 202}:
            message = resp.content.decode()
            try:
                message = json.loads(message)
                message = message["message"]
            except:
                pass
            message = f"Failed to upload to TIRA, got statuscode {resp.status_code}. Details: {message}"
            print(message)
            raise ValueError(message)

        resp = resp.json()
        print(f'Run uploaded to TIRA. Claim ownership via: {self.base_url}/claim-submission/{resp["uuid"]}')

    def create_group(self, vm_id):
        if not vm_id or vm_id != vm_id.lower() or len(vm_id.split()) > 1:
            raise ValueError("The name of the group must be slugified: " + str(vm_id))

        if "tira" in vm_id:
            raise ValueError('The phrase "tira" should not be in the name of the group, got: ' + str(vm_id))

        return self.json_response("/tira-admin/create-group/" + vm_id)

    def register_group(
        self,
        vm_id,
        task_id,
        team_members="",
        name="",
        email="",
        affiliation="",
        country="",
        employment="",
        participates_for="",
        instructor_name="",
        instructor_email="",
        questions="",
    ):

        self.fail_if_api_key_is_invalid()
        user_id = self.json_response("/api/role")["context"]["user_id"]

        endpoint = f"/api/registration/add_registration/{vm_id}/{task_id}"
        body = {
            "group": vm_id,
            "team_members": team_members,
            "team": team_members,
            "registered_on_task": task_id,
            "username": user_id,
            "name": name,
            "email": email,
            "affiliation": affiliation,
            "country": country,
            "employment": employment,
            "participation": participates_for,
            "instructorName": instructor_name,
            "instructorEmail": instructor_email,
            "questions": questions,
        }

        return self.execute_post_return_json(endpoint, json_payload=body)

    def modify_task(self, task_id: str, to_rename: "Dict[str, Any]"):
        task = self.metadata_for_task(task_id)["context"]["task"]
        fields_to_rename = {
            "name": "task_name",
            "description": "task_description",
            "website": "web",
            "help_text": "command_description",
            "help_command": "command_placeholder",
            "task_teams": "allowed_task_teams",
        }
        for k, v in fields_to_rename.items():
            task[k] = task[v]
            del task[v]

        for k, v in to_rename.items():
            assert k in task, k
            task[k] = v

        ret = self.execute_post_return_json("/tira-admin/edit-task/" + task_id, json_payload=task)
        if "status" not in ret or ret["status"] != 0:
            raise ValueError(f"Could not edit task: {ret}")

    def upload_run(
        self,
        file_path: Path,
        dataset_id: str,
        approach: str = None,
        task_id: str = None,
        vm_id: str = None,
        upload_id: str = None,
        allow_multiple_uploads=False,
    ) -> bool:
        """
        Returns true if the upload was successfull, false if not, or none if it was already uploaded.
        """
        logging.info(f"Submitting {upload_id} for Task {task_id}:{dataset_id} on VM {vm_id}")
        if approach:
            task_id, vm_id, display_name = approach.split("/")
            upload_id = self.get_upload_group_id(task_id, vm_id, display_name)

        previous_uploads = self.upload_submissions(task_id, vm_id, upload_id, dataset_id)
        if len(previous_uploads) > 0:
            logging.warn(
                f"Skip upload of file {file_path} for dataset {dataset_id} because there are already"
                f" {len(previous_uploads)} for this dataset. Pass allow_multiple_uploads=True to upload a new dataset"
                " or delete the uploads in the UI."
            )
            return None

        self.fail_if_api_key_is_invalid()

        if file_path is None or not file_path.is_file():
            logging.warn(f"The passed file {file_path} does not exist.")
            raise ValueError(f"The passed file {file_path} does not exist.")

        # TODO: check that task_id and vm_id don't contain illegal characters (e.g., '/')
        url = f"/task/{task_id}/vm/{vm_id}/upload/{dataset_id}/{upload_id}"
        logging.info(f"Submitting the runfile at {url}")
        response = self.execute_post_return_json(url, file_path=file_path)

        return (
            "status" in response
            and response["status"] == 0
            and "message" in response
            and response["message"] == "ok"
            and "new_run" in response
        )

    def get_csrf_token(self):
        self.fail_if_api_key_is_invalid()
        ret = self.json_response("/api/role")
        return ret["csrf"]

    def execute_post_return_json(
        self,
        endpoint: str,
        params: Optional[Union[Dict, List[tuple], bytes]] = None,
        file_path: Path = None,
        json_payload: any = None,
    ) -> Dict:
        assert endpoint.startswith("/")
        csrf = self.get_csrf_token()

        headers = {
            "Api-Key": self.api_key,
            "Api-Username": self.api_user_name,
            "Accept": "application/json",
            "x-csrftoken": csrf,
            "Cookie": f"csrftoken={csrf}",
        }
        for _ in range(self.failsave_retries):
            try:
                files = None if not file_path else {"file": open(file_path, "rb")}

                resp = requests.post(
                    url=f"{self.base_url}{endpoint}", files=files, headers=headers, params=params, json=json_payload
                )
                if resp.status_code not in {200, 202}:
                    raise ValueError(f"Got statuscode {resp.status_code} for {endpoint}. Got {resp.content}")
                else:
                    break
            except Exception as e:
                sleep_time = randint(1, self.failsave_max_delay)
                logging.warn(
                    f"Error occured while fetching {endpoint}. Code: {resp.status_code}. I will sleep"
                    f" {sleep_time} seconds and continue.",
                    exc_info=e,
                )
                time.sleep(sleep_time)

        return resp.json()

    @lru_cache(maxsize=None)
    def archived_json_response(self, endpoint: str, force_reload: bool = False):
        out = Path(self.tira_cache_dir) / ".archived" / Path(endpoint[1:])
        if endpoint.endswith("/"):
            out = out / "index.json"

        if out.exists() and not force_reload:
            return json.load(open(out, "r"))

        out.parent.mkdir(exist_ok=True, parents=True)

        response = self.json_response(endpoint, base_url="https://tira.io")

        with open(out, "w") as f:
            f.write(json.dumps(response))

        return json.load(open(out, "r"))

    @lru_cache(maxsize=None)
    def json_response(self, endpoint: str, params: Optional[Union[Dict, List[tuple], bytes]] = None, base_url=None):
        assert endpoint.startswith("/")
        headers = {"Accept": "application/json"}

        if self.api_key != "no-api-key":
            headers["Api-Key"] = self.api_key
        if self.api_user_name != "no-api-key-user":
            headers["Api-Username"] = self.api_user_name
        base_url = base_url if base_url else self.base_url

        for _ in range(self.failsave_retries):
            try:
                resp = requests.get(url=f"{base_url}{endpoint}", headers=headers, params=params)
                if resp.status_code not in {200, 202}:
                    raise ValueError(f"Got statuscode {resp.status_code} for {endpoint}. Got {resp}")
                else:
                    break
            except Exception as e:
                sleep_time = randint(1, self.failsave_max_delay)
                response_code = "'unknown response code, maybe there was a timeout?'"
                try:
                    response_code = resp.status_code
                except Exception:
                    pass
                logging.warn(
                    f"Error occured while fetching {endpoint}. Code: {response_code}. I will sleep {sleep_time} seconds"
                    " and continue.",
                    exc_info=e,
                )
                time.sleep(sleep_time)

        return resp.json()

    def __listdir_failsave(self, path: str):
        try:
            return os.listdir(path)
        except Exception:
            return []

    def _well_known(self):
        pass

    def input_run_in_sandbox(self, approach: str):
        """
        Returns the directory with the outputs of an approach in mounted into the sandbox. returns None if not in the
        sandbox.
        """
        if "inputRun" not in os.environ:
            return None

        input_run = os.environ["inputRun"]
        input_run_mapping_file = Path(self.tira_cache_dir) / (
            hashlib.md5(input_run.encode("utf-8")).hexdigest()[:6] + "-input-run-mapping.json"
        )

        files_in_input_dir = self.__listdir_failsave(input_run)
        if "1" not in files_in_input_dir or "2" not in files_in_input_dir:
            return input_run

        input_run_mapping = {}
        if input_run_mapping_file.exists():
            input_run_mapping = json.load(open(input_run_mapping_file, "r"))

        if approach not in input_run_mapping:
            next_id = max(list(input_run_mapping.values()) + [0]) + 1
            input_run_mapping[approach] = next_id
            os.makedirs(self.tira_cache_dir, exist_ok=True)
            json.dump(input_run_mapping, open(input_run_mapping_file, "w+"))

        return input_run + "/" + str(input_run_mapping[approach])



import unittest

from tira.rest_api_client import Client


class TestLocalExecutionIntegrationTest(unittest.TestCase):
    def test_pd_load_inputs_wows_re_ranking(self):
        tira = Client()
        actual = tira.pd.inputs("workshop-on-open-web-search", "re-ranking-20231027-training")
        self.assertEqual(6, len(actual))
        self.assertEqual("hubble telescope achievements", actual.iloc[0]["query"])
        self.assertEqual("The Hubble telescope discovered two moons of Pluto, Nix and Hydra.", actual.iloc[0]["text"])
        self.assertEqual(1, actual.iloc[0]["rank"])
        self.assertEqual("1", actual.iloc[0]["qid"])

    def test_pd_load_truths_wows_re_ranking(self):
        tira = Client()
        actual = tira.pd.truths("workshop-on-open-web-search", "re-ranking-20231027-training")
        self.assertEqual(3, len(actual))
        self.assertEqual("hubble telescope achievements", actual.iloc[0]["query"])
        self.assertEqual("1", actual.iloc[0]["qid"])
